"""
The `graph` module provides a wrapper class + convenience methods for
building and querying a Brick graph
"""
import io
from warnings import warn
import os
import sys
import glob
import pkgutil
import importlib.util
import rdflib
import owlrl
import pyshacl
import logging
from typing import List, Optional
from .inference import (
    OWLRLNaiveInferenceSession,
    OWLRLReasonableInferenceSession,
    OWLRLAllegroInferenceSession,
    TagInferenceSession,
    HaystackInferenceSession,
    VBISTagInferenceSession,
)
from . import namespaces as ns

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BrickBase(rdflib.Graph):
    def rebuild_tag_lookup(self, brick_file=None):
        """
        Rebuilds the internal tag lookup dictionary used for Brick tag->class inference.
        This is broken out as its own method because it is potentially an expensive operation.
        """
        self._tagbackend = TagInferenceSession(
            rebuild_tag_lookup=True, brick_file=brick_file, approximate=False
        )

    def to_networkx(self):
        """
        Exports the graph as a NetworkX DiGraph. Edge labels are stored in the 'name' attribute
        Returns:
            graph (networkx.DiGraph): networkx object representing this graph
        """
        try:
            import networkx as nx
        except ImportError as e:
            warn("Could not import NetworkX. Need 'networkx' option during install.")
            raise e
        g = nx.DiGraph()
        for (s, p, o) in self.triples((None, None, None)):
            g.add_edge(s, o, name=p)
        return g

    def get_most_specific_class(self, classlist: List[rdflib.URIRef]):
        """
        Given a list of classes (rdflib.URIRefs), return the 'most specific' classes
        This is a subset of the provided list, containing classes that are not subclasses
        of anything else in the list. Uses the class definitions in the graph to perform
        this task

        Args:
            classlist (list of rdflib.URIRef): list of classes

        Returns:
            classlist (list of rdflib.URIRef): list of specific classes
        """

        # given a list of classes and an ontology (self), return the classes
        # which are not the transtivie parent of any other class in the list.
        specific = []
        for c in classlist:
            # Find the transtive parent by computing transitive closure of rdfs:subClassOf|owl:equivalentClass|brick:aliasOf
            # and then finding the intersection of the closure with the classlist
            # If the intersection is empty, then the class is not a parent of any other class in the list
            # and is therefore specific
            # if the intersection is only the class itself or anything it is equivalent to, then it is specific
            closure_query = f"""
            SELECT ?parent WHERE {{
                ?parent (rdfs:subClassOf|owl:equivalentClass)+ <{c}> .
            }}
            """
            closure = set(x[0] for x in self.query(closure_query))

            equivalent_query = f"""
            SELECT ?parent WHERE {{
                <{c}> (owl:equivalentClass|brick:aliasOf)+ ?parent .
            }}
            """
            equivalent = set(x[0] for x in self.query(equivalent_query))

            if len(closure.intersection(classlist)) == 0 or closure.intersection(
                classlist
            ).issubset(equivalent):
                specific.append(c)

        return specific

    def validate(
        self,
        extra_graphs: Optional[List[rdflib.Graph]] = None,
        engine: Optional[str] = None,
        min_iterations=1,
        max_iterations=10,
    ):
        """
        Validates the graph using the shapes embedded w/n the graph.

        Args:
          extra_graphs (list of rdflib.Graph or brickschema.graph.Graph): merges these graphs and includes them in
                the validation
          engine (str): the SHACL engine to use. Options are 'pyshacl' and 'topquadrant'. Defaults to 'topquadrant'
                if available, else 'pyshacl'.
          min_iterations (int): minimum number of iterations for topquadrant engine.
          max_iterations (int): maximum number of iterations for topquadrant engine.

        Returns:
          (conforms, resultsGraph, resultsText) from pyshacl
        """
        shapes = rdflib.Graph()
        if extra_graphs is not None and isinstance(extra_graphs, list):
            for sg in extra_graphs:
                shapes += sg

        if engine is None:
            if importlib.util.find_spec("brick_tq_shacl") is not None:
                engine = "topquadrant"
            else:
                engine = "pyshacl"

        if engine == "pyshacl":
            if min_iterations > 1 or max_iterations > 1:
                self.compile(engine="pyshacl", min_iterations=min_iterations, max_iterations=max_iterations)
            combined = self.skolemize() + shapes
            return pyshacl.validate(
                combined,
                advanced=True,
                abort_on_first=True,
                allow_warnings=True,
            )
        elif engine == "topquadrant":
            from brick_tq_shacl import validate

            return validate(
                self,
                shapes,
                min_iterations=min_iterations,
                max_iterations=max_iterations,
            )

    def serve(self, address="127.0.0.1:8080", ignore_prefixes=[]):
        """
        Start web server offering SPARQL queries and 1-click reasoning capabilities

        Args:
          address (str): <host>:<port> of the web server
          ignore_prefixes (list[str]): list of prefixes not to be added to the query editor's namespace bindings.
        """
        try:
            from . import web
        except ImportError:
            warn(
                "Using the webserver requires the 'web' option:\n\n\tpip install brickschema[web]"
            )
            import sys

            sys.exit(1)
        srv = web.Server(self, ignore_prefixes=ignore_prefixes)
        srv.start(address)

    def get_extensions(self):
        """
        Returns a list of Brick extensions

        This currently just lists the extensions already loaded into brickschema,
        but may in the future pull a list of extensions off of an online resolver
        """
        d = os.path.dirname(sys.modules[__name__].__file__)
        extension_path = os.path.join(
            d, "ontologies", self._brick_version, "extensions"
        )
        extensions = glob.glob(os.path.join(extension_path, "*.ttl"))
        return [
            os.path.basename(x).strip(".ttl")[len("brick_extension_") :]
            for x in extensions
        ]

    def get_alignments(self):
        """
        Returns a list of Brick alignments

        This currently just lists the alignments already loaded into brickschema,
        but may in the future pull a list of alignments off of an online resolver
        """
        d = os.path.dirname(sys.modules[__name__].__file__)
        alignment_path = os.path.join(
            d, "ontologies", self._brick_version, "alignments"
        )
        alignments = glob.glob(os.path.join(alignment_path, "*.ttl"))
        return [
            os.path.basename(x)[len("Brick-") : -len("-alignment.ttl")]
            for x in alignments
        ]

    def compile(
        self,
        extra_graphs: Optional[List[rdflib.Graph]] = None,
        engine=None,
        min_iterations=1,
        max_iterations=10,
    ):
        """
        Compiles the graph by applying SHACL-AF rules. This includes tag inference
        if the tag inference rules are loaded into the graph.

        Possible engines are:
        - 'pyshacl': default, python-based SHACL implementation
        - 'topquadrant': uses TopQuadrant's SHACL-AF implementation. Requires 'brick-tq-shacl'
          to be installed.

        Args:
            extra_graphs (list[Graph]): list of graphs containing extra ontological definitions. If not provided,
                uses the ontologies loaded in the graph.
            engine (str): which SHACL engine to use. If not provided, defaults to topquadrant
                then pyshacl
            min_iterations (int): minimum number of iterations for pyshacl or topquadrant engine.
            max_iterations (int): maximum number of iterations for pyshacl or topquadrant engine.
        """
        onts = rdflib.Graph()
        if extra_graphs:
            for g in extra_graphs:
                onts += g

        shacl_engine = engine
        if shacl_engine is None:
            if importlib.util.find_spec("brick_tq_shacl") is not None:
                shacl_engine = "topquadrant"
            else:
                shacl_engine = "pyshacl"

        if shacl_engine == "topquadrant":
            try:
                from brick_tq_shacl import infer as tq_shacl_infer

                res = tq_shacl_infer(
                    self,
                    onts,
                    min_iterations=min_iterations,
                    max_iterations=max_iterations,
                )
                self += res
                return self
            except ImportError:
                warn(
                    "TopQuadrant SHACL engine selected/defaulted, but failed to import. Falling back to pyshacl."
                )
                shacl_engine = "pyshacl"

        if shacl_engine == "pyshacl":
            if max_iterations < min_iterations:
                max_iterations = min_iterations
            skolemized = self.skolemize()
            combined = skolemized + onts
            for i in range(max_iterations):
                old_size = len(combined)
                valid, _, report = pyshacl.validate(
                    data_graph=combined,
                    advanced=True,
                    allow_warnings=True,
                    abort_on_first=True,
                    inplace=True,
                )
                if not valid:
                    warn(report)
                if (i + 1) >= min_iterations and len(combined) == old_size:
                    break
            added = combined - skolemized - onts
            self += added
            return self
        raise Exception(f"Unknown SHACL engine {engine}")

    def expand(
        self,
        profile,
        backend=None,
        simplify=True,
    ):
        """
        Expands the current graph with the inferred triples under the given entailment regime
        and with the given backend. Possible profiles are:
        - 'rdfs': runs RDFS rules
        - 'owlrl': runs full OWLRL reasoning
        - 'vbis': adds VBIS tags

        Possible backends are:
        - 'reasonable': default, fastest backend
        - 'allegrograph': uses Docker to interface with allegrograph
        - 'owlrl': native-Python implementation

        Not all backend work with all profiles. In that case, brickschema will use the fastest appropriate
        backend in order to perform the requested inference.

        To perform more than one kind of inference in sequence, use '+' to join the profiles:

            import brickschema
            g = brickschema.Graph()
            g.expand(profile='rdfs+owlrl') # performs RDFS inference, then OWLRL inference


        # TODO: currently nothing is cached between expansions
        """
        if "+" in profile:
            for prf in profile.split("+"):
                self.expand(
                    prf,
                    backend=backend,
                    simplify=simplify,
                )
            return

        if profile == "brick":
            return self.expand("owlrl", backend=backend, simplify=simplify)
        elif profile == "rdfs":
            owlrl.DeductiveClosure(owlrl.RDFS_Semantics).expand(self)
            return
        elif profile == "owlrl":
            if backend is None:
                backend = "reasonable"
            if backend == "reasonable":
                OWLRLReasonableInferenceSession().expand(self)
            elif backend == "allegrograph":
                OWLRLAllegroInferenceSession().expand(self)
            elif backend == "owlrl":
                OWLRLNaiveInferenceSession().expand(self)
            else:
                raise ValueError(f"Unknown owlrl backend {backend}")
        elif profile == "vbis":
            VBISTagInferenceSession(brick_version=self._brick_version).expand(self)
        else:
            raise Exception(f"Invalid profile '{profile}'")

        if simplify:
            self.simplify()
        return self

    def simplify(self):
        """
        Removes redundant and axiomatic triples and other detritus that is produced as a side effect of reasoning.
        Simplification consists of the following steps:
        - remove all "a owl:Thing", "a owl:Nothing" triples
        - remove all "a <blank node" triples
        - remove all "X owl:sameAs Y" triples
        """
        for entity, etype in self.subject_objects(ns.RDF.type):
            if etype in [ns.OWL.Thing, ns.OWL.Nothing]:
                self.remove((entity, ns.A, etype))
            elif isinstance(etype, rdflib.BNode):
                self.remove((entity, ns.A, etype))

        for a, b in self.subject_objects(ns.OWL.sameAs):
            if a == b:
                self.remove((a, ns.OWL.sameAs, b))


class GraphCollection(rdflib.Dataset, BrickBase):
    def __init__(
        self,
        *args,
        load_brick=False,
        load_brick_nightly=False,
        brick_version="1.4",
        **kwargs,
    ):
        """Wrapper class and convenience methods for handling Brick models
        and graphs. Accepts the same arguments as RDFlib.Graph

        Args:
            load_brick (bool): if True, loads packaged Brick ontology
                into graph
            load_brick_nightly (bool): if True, loads latest nightly Brick build
                into graph (requires internet connection)
            brick_version (string): the MAJOR.MINOR version of the Brick ontology
                to load into the graph. Only takes effect for the load_brick argument

        Returns:
            A GraphCollection object
        """
        kwargs.update({"default_union": True})
        super().__init__(*args, **kwargs)
        self._brick_version = brick_version
        self._load_brick = load_brick
        self._load_brick_nightly = load_brick_nightly
        self._graph_init()
        # subset of graphs in the store to use; if this is length-0, then
        # all graphs are used
        self._subset = set()

    def __iter__(self):
        """Iterates over all quads in the store"""
        for (s, p, o, _) in self.quads((None, None, None, None)):
            yield (s, p, o)

    def load_graph(
        self,
        filename: str = None,
        source: io.IOBase = None,
        format: str = None,
        graph: rdflib.Graph = None,
        graph_name: rdflib.URIRef = None,
    ):
        """
        Imports the triples contained in the indicated file (or graph) into the graph.
        Names the graph using any owl:Ontology declaration found in the file
        or using the 'graph_name' argument if it is provided

        Args:
            filename (str): relative or absolute path to the file
            source (file): file-like object
            graph (brickschema.Graph): graph to load into the collection
            graph_name (rdflib.URIRef): name of the graph (defaults to owl:Ontology instance or 'default')

        Return:
            parsed (rdflib.Graph): the graph loaded from parsing the input
        """
        if graph is None:
            graph = Graph().load_file(filename=filename, source=source, format=format)
        if graph_name is None:
            try:
                graph_name = next(graph.subjects(rdflib.RDF.type, ns.OWL.Ontology))
            except StopIteration:
                warn(
                    f"No owl:Ontology found in graph {filename or source}. Using default graph"
                )
                graph_name = rdflib.graph.DATASET_DEFAULT_GRAPH_ID
        else:
            graph_name = rdflib.URIRef(graph_name)
        g = self.graph(graph_name)
        for (s, p, o) in graph.triples((None, None, None)):
            g.add((s, p, o))
        return g

    def remove_graph(self, graph_name):
        """
        Removes the named graph from the graph store

        Args:
            graph_name (str): name of the graph to remove
        """
        self.remove_graph(graph_name)

    def _graph_init(self):
        """
        Initializes the graph by downloading or loading from local cache the requested
        versions of the Brick ontology.
        """
        ns.bind_prefixes(self, brick_version=self._brick_version)

        if self._load_brick_nightly:
            self.load_graph(
                filename="https://github.com/BrickSchema/Brick/releases/download/nightly/Brick.ttl",
                format="turtle",
                graph_name="https://brickschema.org/schema/Brick#",
            )
        elif self._load_brick:
            # get ontology data from package
            data = pkgutil.get_data(
                __name__, f"ontologies/{self._brick_version}/Brick.ttl"
            ).decode()
            # wrap in StringIO to make it file-like
            self.load_graph(
                source=io.StringIO(data),
                format="turtle",
                graph_name="https://brickschema.org/schema/Brick#",
            )

        self._tagbackend = None

    @property
    def graph_names(self):
        """
        Returns a list of the names of the graphs in the graph store

        Returns:
            list: list of graph names
        """
        if self._subset:
            return list(self._subset)
        else:
            return [g.identifier for g in self.graphs()]

    def load_alignment(self, alignment_name):
        """
        Loads the given alignment into the current graph by name.
        Use get_alignments() to get a list of alignments
        """
        alignment_name = f"Brick-{alignment_name}-alignment.ttl"
        alignment_path = os.path.join(
            "ontologies", self._brick_version, "alignments", alignment_name
        )
        data = pkgutil.get_data(__name__, alignment_path).decode()
        # wrap in StringIO to make it file-like
        self.load_graph(source=io.StringIO(data), format="turtle")

    def load_extension(self, extension_name):
        """
        Loads the given extension into the current graph by name.
        Use get_extensions() to get a list of extensions
        """
        extension_name = f"brick_extension_{extension_name}.ttl"
        extension_path = os.path.join(
            "ontologies", self._brick_version, "extensions", extension_name
        )
        data = pkgutil.get_data(__name__, extension_path).decode()
        # wrap in StringIO to make it file-like
        self.load_graph(source=io.StringIO(data), format="turtle")

    def contexts(self, triple=None):
        """Iterate over all contexts in the graph

        If triple is specified, iterate over all contexts the triple is in.
        """
        for context in self.store.contexts(triple):
            if len(self._subset) > 0 and context not in self._subset:
                continue
            if isinstance(context, Graph):
                # TODO: One of these should never happen and probably
                # should raise an exception rather than smoothing over
                # the weirdness - see #225
                yield context
            else:
                yield self.get_context(context)


class Graph(BrickBase):
    def __init__(
        self,
        *args,
        load_brick=False,
        load_brick_nightly=False,
        brick_version="1.4",
        _delay_init=False,
        **kwargs,
    ):
        """Wrapper class and convenience methods for handling Brick models
        and graphs. Accepts the same arguments as RDFlib.Graph

        Args:
            load_brick (bool): if True, loads packaged Brick ontology
                into graph
            load_brick_nightly (bool): if True, loads latest nightly Brick build
                into graph (requires internet connection)
            brick_version (string): the MAJOR.MINOR version of the Brick ontology
                to load into the graph. Only takes effect for the load_brick argument
            _delay_init (bool): if True, the graph will not call internal initialization logic.
                You should not need to touch this.

        Returns:
            A Graph object
        """
        super().__init__(*args, **kwargs)
        self._brick_version = brick_version
        self._load_brick = load_brick
        self._load_brick_nightly = load_brick_nightly
        if not _delay_init:
            self._graph_init()

    def _graph_init(self):
        """
        Initializes the graph by downloading or loading from local cache the requested
        versions of the Brick ontology. If Graph() is initialized Using Graph.open() as
        part of an external store will also call this method automatically
        """
        ns.bind_prefixes(self, brick_version=self._brick_version)

        if self._load_brick_nightly:
            self.load_file(
                "https://github.com/BrickSchema/Brick/releases/download/nightly/Brick.ttl",
                format="turtle",
            )
        elif self._load_brick:
            # get ontology data from package
            data = pkgutil.get_data(
                __name__, f"ontologies/{self._brick_version}/Brick.ttl"
            ).decode()
            # wrap in StringIO to make it file-like
            self.load_file(source=io.StringIO(data), format="turtle")

        self._tagbackend = None

    def load_file(self, filename=None, source=None, format=None):
        """
        Imports the triples contained in the indicated file into the default graph.

        Args:
            filename (str): relative or absolute path to the file
            source (file): file-like object
        """
        if filename is not None:
            fmt = format if format else rdflib.util.guess_format(filename)
            self.parse(filename, format=fmt)
        elif source is not None:
            for fmt in [format, "ttl", "n3", "xml"]:
                try:
                    self.parse(source=source, format=fmt)
                    return self
                except Exception as e:
                    warn(f"could not load {filename} as {fmt}: {e}")
            raise Exception(f"unknown file format for {filename}")
        else:
            raise Exception(
                "Must provide either a filename or file-like\
source to load_file"
            )
        return self

    def add(self, *triples):
        """
        Adds triples to the graph. Triples should be 3-tuples of rdflib.Nodes.

        If the object of a triple is a list/tuple of length-2 lists/tuples,
        then this method will substitute a blank node as the object of the original
        triple, add the new triples, and add as many triples as length-2 items in the
        list with the blank node as the subject and the item[0] and item[1] as the predicate
        and object, respectively.

        For example, calling add((X, Y, [(A,B), (C,D)])) produces the following triples::

            X Y _b1 .
            _b1 A B .
            _b1 C D .

        or, in turtle::

            X Y [
              A B ;
              C D ;
            ] .

        Otherwise, acts the same as rdflib.Graph.add
        """
        for triple in triples:
            assert len(triple) == 3
            obj = triple[2]
            if isinstance(obj, (list, tuple)):
                for suffix in obj:
                    assert len(suffix) == 2
                bnode = rdflib.BNode()
                self.add((triple[0], triple[1], bnode))
                for (nested_pred, nested_obj) in obj:
                    self.add((bnode, nested_pred, nested_obj))
            else:
                super().add(triple)

    @property
    def nodes(self):
        """
        Returns all nodes in the graph

        Returns:
            nodes (list of rdflib.URIRef): nodes in the graph
        """
        return self.all_nodes()

    def from_haystack(self, namespace, model):
        """
        Adds to the graph the Brick triples inferred from the given Haystack model.
        The model should be a Python dictionary produced from the Haystack JSON export

        Args:
            model (dict): a Haystack model
        """
        sess = HaystackInferenceSession(namespace)
        self.add(*sess.infer_model(model))
        return self

    def from_triples(self, triples):
        """
        Creates a graph from the given list of triples

        Args:
            triples (list of rdflib.Node): triples to add to the graph
        """
        self.add(*triples)
        return self

    def load_alignment(self, alignment_name):
        """
        Loads the given alignment into the current graph by name.
        Use get_alignments() to get a list of alignments
        """
        alignment_name = f"Brick-{alignment_name}-alignment.ttl"
        alignment_path = os.path.join(
            "ontologies", self._brick_version, "alignments", alignment_name
        )
        data = pkgutil.get_data(__name__, alignment_path).decode()
        # wrap in StringIO to make it file-like
        self.load_file(source=io.StringIO(data), format="turtle")

    def load_extension(self, extension_name):
        """
        Loads the given extension into the current graph by name.
        Use get_extensions() to get a list of extensions
        """
        extension_name = f"brick_extension_{extension_name}.ttl"
        extension_path = os.path.join(
            "ontologies", self._brick_version, "extensions", extension_name
        )
        data = pkgutil.get_data(__name__, extension_path).decode()
        # wrap in StringIO to make it file-like
        self.load_file(source=io.StringIO(data), format="turtle")



"""
Python package `brickschema` provides a set of tools, utilities and interfaces
for working with, developing and interacting with Brick models.
"""

import logging
from . import inference, namespaces, graph
from .graph import Graph, GraphCollection

logging.basicConfig(
    format="%(asctime)s,%(msecs)03d %(levelname)-7s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=logging.WARNING,
)

has_sqlalchemy = False
try:
    import rdflib_sqlalchemy
    has_sqlalchemy = True
except ImportError as e:
    print(e)
    logging.warning(
        "sqlalchemy not installed. SQL-backed graph support will not be available. Try 'pip install brickschema[persistence]' to install it."
    )

__version__ = "0.2.0"
__all__ = ["graph", "inference", "namespaces"]
if has_sqlalchemy:
    __all__.append("rdflib_sqlalchemy")



from typing import Optional

import importlib_resources
import rdflib

from brickschema.brickify.src.handlers.Handler.Handler import Handler
from brickschema.brickify.src.handlers.Handler.HaystackHandler.utils.HaystackRDFInferenceSession import (
    HaystackRDFInferenceSession,
)


class HaystackHandler(Handler):
    def __init__(
        self,
        source,
        input_format: Optional[str] = "turtle",
        config_file: Optional[str] = None,
    ):
        """
        The HaystackHandler is used to convert Haystack v4 graphs to brick graphs. The HaystackHandler
        loads in the HaystackDefinitions and an analogy data (graph) during initialization.

        :param source: A filepath/URL
        :param input_format:  Input format of the file
        :param config_file: Custom conversion configuration file
        """
        module_path = (
            [
                "brickschema.brickify.src.handlers.Handler.HaystackHandler.conversions",
                "haystack.json",
            ]
            if not config_file
            else []
        )
        super().__init__(
            source=source,
            input_format=input_format,
            module_path=module_path,
            config_file=config_file,
        )
        self.hs_graph = rdflib.Graph()
        self.h2b_graph = rdflib.Graph()
        self.hs_graph.parse(
            source="https://project-haystack.org/download/defs.ttl", format="turtle"
        )
        with importlib_resources.path(module_path[0], "analogy.ttl") as data_file:
            with open(data_file, "r") as h2b:
                self.h2b_graph.parse(h2b, format="turtle")

    def ingest_data(self):
        """
        Extends the ingest_data() method to append Haystack definitions and analogy graphs to the output graph.
        The analogy data is used in the translation step (example: phIoT:submeterOf is similar to brick:isPartOf).
        """
        super().ingest_data()
        self.graph += self.hs_graph
        self.graph += self.h2b_graph

    def infer(self):
        """
        Uses HaystackRDFInferenceSession to extract tags from instance types (classes) and use a
        TagInferenceSession to provide inference of a Brick model from a Haystack model.
        """
        super().infer()
        haysess = HaystackRDFInferenceSession("https://project-haystack.org/example#")
        haysess.infer_model(self.graph)

    def clean_up(self):
        """
        Removes the analogy graph and the haystack definitions, haystack triples from the output graph.
        """
        self.graph -= self.hs_graph
        self.graph -= self.h2b_graph
        self.graph.update(
            'DELETE { ?subject ?predicate ?object . } WHERE { ?subject ?predicate ?object . FILTER ( STRSTARTS(STR(?predicate), "https://project-haystack.org/def/ph") )  }'
        )
        self.graph.update(
            'DELETE { ?subject ?predicate ?object . } WHERE { ?subject ?predicate ?object . FILTER ( STRSTARTS(STR(?subject), "https://project-haystack.org/def/ph") )  }'
        )
        self.graph.update(
            'DELETE { ?subject ?predicate ?object . } WHERE { ?subject ?predicate ?object . FILTER ( STRSTARTS(STR(?object), "https://project-haystack.org/def/ph") )  }'
        )
        self.graph.update(
            """
            DELETE { ?part brick:hasLocation ?location . }
            WHERE { ?thing brick:hasLocation ?location . ?part brick:isPartOf|^brick:hasPart ?thing . }
            """
        )
        self.graph.update(
            """
            DELETE { ?thing brick:hasPart ?part . ?part brick:isPartOf ?thing . }
            WHERE { ?part brick:isPointOf|^brick:hasPoint ?thing . ?part brick:isPartOf|^brick:hasPart ?thing . }
            """
        )



import pytest
from ontoenv import OntoEnv, Config
import brickschema
from rdflib import RDF, RDFS, BRICK, OWL, Namespace

QUDT = Namespace("http://qudt.org/schema/qudt/")

# using code from https://docs.pytest.org/en/latest/example/simple.html


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def pytest_generate_tests(metafunc):
    """
    Generates Brick tests for a variety of contexts
    """

    # validates that example files pass validation
    if "owlrl_inference_backend" in metafunc.fixturenames:
        metafunc.parametrize("owlrl_inference_backend", ["owlrl", "allegrograph", "reasonable"])


@pytest.fixture()
def brick_with_imports():
    cfg = Config([], strict=False, offline=False, temporary=True)
    env = OntoEnv(cfg)
    # TODO: need to add rdflib graph to the environment directly
    g = brickschema.Graph(load_brick=True)
    g.bind("qudt", QUDT)
    g.bind("rdf", RDF)
    g.bind("rdfs", RDFS)
    g.bind("brick", BRICK)
    imported = env.import_dependencies(g, fetch_missing=True, recursion_depth=1)
    print(f"Imported {len(imported)} dependencies into the Brick graph.: {imported}")
    g.serialize("/tmp/brick_with_imports.ttl", format="turtle")
    return g


@pytest.fixture(params=["pyshacl", "topquadrant"])
def shacl_engine(request):
    """
    Parametrizes tests over both SHACL engines.

    Skips TopQuadrant engine if brick_tq_shacl is not installed.
    """
    if request.param == "topquadrant":
        import importlib.util

        if importlib.util.find_spec("brick_tq_shacl") is None:
            pytest.skip("brick_tq_shacl not installed; skipping TopQuadrant engine tests")
    return request.param



from pathlib import Path
from typing import Optional

import typer
from tabulate import tabulate

from brickschema.brickify.src.handlers.Handler.TableHandler import TableHandler
from brickschema.brickify.util import (
    get_workbook,
    find_header_row,
    cleaned_value,
    ignore_row,
)


class RACHandler(TableHandler):
    def __init__(
        self,
        source,
        input_format: Optional[str] = "xls",
        config_file: Optional[str] = None,
    ):
        """
        RACHandler is a TableHandler designed to work on Excel Workbooks (only overrides the ingestion method).
        The default template works with MetaSys RAC schedules.

        :param source: A filepath
        :param input_format: Input format (.xls)
        :param config_file: Custom conversion configuration file
        """
        module_path = (
            [
                "brickschema.brickify.src.handlers.Handler.RACHandler.conversions",
                "rac.yml",
            ]
            if not config_file
            else []
        )
        super().__init__(
            source=source,
            input_format=input_format,
            module_path=module_path,
            config_file=config_file,
        )

    def ingest_data(self):
        """
        Splits the workbook data into multiple sheets and stores the rows in a key-value based
        data model where the key is the cell's column header, and the value is the cell's cleaned value.
        """
        workbook = get_workbook(Path(self.source))
        sheets = workbook.sheets()
        table = [(index, sheet.name) for index, sheet in enumerate(sheets)]
        typer.echo(
            tabulate(table, headers=["Sheet ID", "Sheet Name"], tablefmt="pretty")
        )
        sheet_ids = typer.prompt(text="Enter the Sheet IDs", default="all")
        if sheet_ids == "all":
            sheet_ids = range(len(sheets))
        else:
            sheet_ids = [int(sheet_id.strip()) for sheet_id in sheet_ids.split(",")]
            invalid_sheet_ids = [
                str(sheet_id)
                for sheet_id in sheet_ids
                if sheet_id not in range(len(sheets))
            ]
            if invalid_sheet_ids:
                typer.echo(
                    typer.style(
                        f"[INFO] Skipping invalid sheet IDs: {', '.join(invalid_sheet_ids)}",
                        fg=typer.colors.YELLOW,
                    )
                )

        for index, sheet in enumerate(sheets):
            if index not in sheet_ids:
                continue
            header_row = find_header_row(
                sheet=sheet, header_start=self.config["header_start"]
            )
            if not header_row:
                continue
            header = sheet.row_values(header_row)
            header = cleaned_value(
                value=header, replace_dict=self.config["replace_dict"]["headers"]
            )
            header_dict = {idx: value for idx, value in enumerate(header)}
            for row_number in range(header_row + 1, sheet.nrows):
                row = sheet.row_values(row_number)
                if ignore_row(row):
                    continue
                else:
                    row_object = {
                        header_dict[key]: value for key, value in enumerate(row)
                    }
                    update_dict = {}
                    for key, value in row_object.items():
                        if "/" in key:
                            keys = key.split("/")
                            values = value.split("/")
                            for idx, k in enumerate(keys):
                                if len(values) > idx:
                                    update_dict[k] = values[idx]
                                else:
                                    update_dict[k] = None
                    row_object = {**row_object, **update_dict}
                    if row_object:
                        row_object = {
                            key: cleaned_value(
                                value,
                                replace_dict=self.config["replace_dict"]["values"],
                            )
                            for key, value in row_object.items()
                        }
                        self.data.append(row_object)



import csv
import re
import traceback
from pathlib import Path
from typing import Optional, List

from jinja2 import Template
from typer import progressbar

from brickschema.brickify.src.handlers.Handler.Handler import Handler
from brickschema.brickify.util import cleaned_value


class TableHandler(Handler):
    def __init__(
        self,
        source,
        input_format: Optional[str] = "csv",
        module_path: Optional[List[str]] = None,
        config_file: Optional[Path] = None,
    ):
        """
        A handler designed specifically to deal with tabular data (overrides ingestion and translations methods only).

        :param source: A filepath
        :param input_format: Input format of the file (supports csv, tsv by default)
        :param module_path: Path to default template files in the package ([<dot-separate-module-path>, <template-filename>])
        :param config_file: Custom conversion configuration file
        """
        super().__init__(
            source=source,
            input_format=input_format,
            module_path=module_path,
            config_file=config_file,
        )
        self.data = []
        self.dialect = "excel"
        if input_format == "tsv":
            self.dialect = "excel-tab"

    def ingest_data(self):
        """
        Ingests tabular data into a key-value based data model where the key is the column header, and value is
        the cleaned cell value.
        """
        with open(self.source, newline="") as csv_file:
            reader = csv.DictReader(csv_file, dialect=self.dialect)
            for row in reader:
                replace_dict = (
                    self.config["replace_dict"]["values"]
                    if "replace_dict" in self.config
                    and "values" in self.config["replace_dict"]
                    else {}
                )
                item = {
                    key.strip(): cleaned_value(
                        value,
                        replace_dict=replace_dict,
                    )
                    for key, value in row.items()
                }
                self.data.append(item)

    def translate(self):
        """
        Translates the ingested flattened data to triples, into the output graph.
        """
        if "macros" in self.config:
            macros = "\n".join(self.config["macros"])
        else:
            macros = ""
        with progressbar(self.data) as data:
            for item in data:
                query = None
                for operation in self.config["operations"]:
                    args_finder = re.compile(r"{([^(?!{}).*$]*?)\}")
                    if "template" in operation:
                        template_string = operation["template"]
                        if macros:
                            template_string = macros + "\n" + template_string
                        template = Template(template_string)
                        query = f"INSERT DATA {{{{ {template.render(value=item)} }}}}"
                    elif "data" in operation:
                        query = f"INSERT DATA {{{{ {operation['data']} }}}}"
                    elif "query" in operation:
                        query = operation["query"]
                    if not query:
                        continue
                    args = args_finder.findall(query)
                    args = [arg.strip() for arg in args if arg.strip()]
                    args_list = list(set([arg.strip() for arg in args])) or []
                    if "conditions" in operation:
                        try:
                            conditions = [
                                eval(condition.format_map(item))
                                for condition in operation["conditions"]
                            ]
                            conditions = cleaned_value(conditions)
                        except KeyError as e:
                            print(e)
                            traceback.print_exc()
                            conditions = [False]
                    else:
                        conditions = []
                    if all([arg in item.keys() for arg in args_list] + conditions):
                        query_str = query.format_map(item)
                        try:
                            self.graph.update(query_str)
                        except Exception as e:
                            print(e)
                            print(query_str)
                            traceback.print_exc()
                            exit(1)



"""
The util module provides helper functions used by brickify.
"""

import json
import re
from pathlib import Path
from typing import Optional, Dict

import click_spinner
import typer
import yaml
from rdflib import Namespace, OWL, RDF, RDFS, Graph
from xlrd import open_workbook


def cleaned_value(value, replace_dict: Optional[Dict] = {}):
    """
    Returns a cleaned value produced by doing regex replacements and elimination
    of leading or trailing whitespaces.

    :param value: List|float|string
    :param replace_dict: Key-value pairs for regex replacements
    :returns: cleaned List|float|string
    """
    if isinstance(value, float):
        return int(value) if str(value)[-2:] == ".0" else value
    if isinstance(value, list):
        return [cleaned_value(item, replace_dict) for item in value]
    clean_value = value
    if isinstance(value, str):
        try:
            if "." in value:
                return float(value)
            return int(value)
        except ValueError:
            pass
        if value in ["TRUE", "true", "True", "on", "ON"]:
            return True
        if value in ["FALSE", "false", "False", "off", "OFF"]:
            return False
        for replacement in replace_dict.items():
            clean_value = re.sub(*replacement, clean_value)
        return clean_value.strip()
    return clean_value


def get_workbook(filename: Path):
    """

    :param filename: Input filepath
    :returns: An XLRD Workbook object
    """
    if not filename.is_file():
        message_start = typer.style("[Error] Input file: ", fg=typer.colors.RED)
        filename = typer.style(f"{filename}", fg=typer.colors.RED, bold=True)
        message_end = typer.style(" does not exist!", fg=typer.colors.RED)
        typer.echo(message_start + filename + message_end)
        raise typer.Exit(code=1)
    else:
        try:
            workbook = open_workbook(filename=filename)
        except Exception:
            message_start = typer.style("[Error] Input file", fg=typer.colors.RED)
            filename = typer.style(f"{filename}", fg=typer.colors.RED, bold=True)
            message_end = typer.style(
                " has an unsupported format!", fg=typer.colors.RED
            )
            typer.echo(message_start + filename + message_end)
            raise typer.Exit(code=1)
        return workbook


def find_header_row(sheet, header_start):
    """
    Finds the header row number in a sheet of an XLRD Workbook.

    :param sheet: Input sheet
    :param header_start: Header pattern (a substring of the first header)
    :return: Row number for the header row
    """
    row_number = 0
    for row in sheet.get_rows():
        if header_start in str(row[0]):
            return row_number
        row_number += 1
    return None


def is_title_row(row):
    return is_empty_row(row[1:])


def is_empty_row(row):
    return not any(row)


def is_not_data(row):
    """

    :param row: Input row
    :return: True
    """
    return is_empty_row(row[:2])


def ignore_row(row):
    """
    Finds out if a row should be ignored. based based on the presence of data in all the
    required table columns.

    :param row: Input row
    :return: True if the row is a title row or doesn't have data, otherwise False
    """
    return is_title_row(row) or is_not_data(row) or is_empty_row(row)


def not_important(row, required_fields):
    """
    Finds out if a row is important based based on the presence of data in all the
    required table columns.

    :param row: Input row
    :param required_fields: List of indices of required fields
    :rtype: bool
    """
    return all([value for index, value in enumerate(row) if index in required_fields])


def get_required(header):
    """
    Returns a list of column indices that contain the substring "required".

    :param header: List of column headers
    :return: List of column indices
    """
    return [
        index for index, value in enumerate(header) if "(required)" in value.lower()
    ]


def bind_namespaces(graph, namespace_prefixes: Dict[str, str]):
    """
    Binds namespace prefixes to an rdflib Graph.

    :param graph: Input graph (rdflib.Graph)
    :param namespace_prefixes: A dictionary of key-value pairs {"prefixA": "namespaceA", ...}
    """
    graph.bind("rdf", RDF)
    graph.bind("rdfs", RDFS)
    graph.bind("owl", OWL)
    for prefix, namespace in namespace_prefixes.items():
        graph.bind(prefix, Namespace(namespace))


def minify_graph(graph, brick_file):
    """
    Compresses the output graph by removing inferable triples.

    :param graph: Input graph (rdflib.Graph)
    :param brick_file: Brick.ttl filepath/URL to use as a reference
    """
    original = len(graph)
    brick = Graph()
    if not brick_file:
        typer.echo(
            typer.style(
                "[INFO] Loading the latest nightly release of brick",
                fg=typer.colors.YELLOW,
            )
        )
        with click_spinner.spinner():
            brick.load(
                "https://github.com/BrickSchema/Brick/releases/download/nightly/Brick.ttl",
                format="turtle",
            )
    else:
        brick.load(str(brick_file), format="turtle")
    graph += brick
    graph.update(
        """
    DELETE {
        ?instance a ?type .
    }
    WHERE {
        ?instance a ?type .
        ?instance a ?subtype .
        ?subtype (rdfs:subClassOf|owl:equivalentClass)+ ?type .
    }
    """
    )
    graph -= brick
    minified = len(graph)
    typer.echo(
        typer.style(
            f"[INFO] Inferable triples removed: {original - minified} ",
            fg=typer.colors.GREEN,
        )
    )
    typer.echo(
        typer.style(
            f"[INFO] Total triples generated: {minified}",
            fg=typer.colors.GREEN,
        )
    )


def load_config(fp, filename: str):
    """
    Parses and returns the conversion configuration from JSON/YAML files.

    :param fp: file pointer
    :param filename: filename (used to identify which parser to use)

    :returns: dict
    """
    filename = str(filename)
    if filename.endswith(".json"):
        return json.load(fp)
    elif filename.endswith(".yml"):
        return yaml.load(fp, Loader=yaml.FullLoader)
    else:
        typer.echo(
            typer.style(
                "[WARN] '.json' and '.yml' configurations are supported.",
                fg=typer.colors.YELLOW,
            )
        )
        return {}



from brickschema.namespaces import BRICK, A, TAG
from rdflib import Namespace
from brickschema.persistent import PersistentGraph


def test_persistent_graph(shacl_engine):
    pg = PersistentGraph("sqlite://")
    assert len(pg) == 0

    pg = PersistentGraph("sqlite://", load_brick=True)
    assert len(pg) > 0

    EX = Namespace("http://example.com/building#")

    pg.add((EX["a"], A, BRICK.Temperature_Sensor))
    pg.compile(engine=shacl_engine)
    pg.serialize("/tmp/out.ttl", format="turtle")
    assert (EX["a"], BRICK.hasTag, TAG.Temperature) in pg

    res = pg.query("SELECT * WHERE { ?x a brick:Temperature_Sensor }")
    assert len(res) == 1

    pg2 = PersistentGraph("sqlite://", load_brick=False)
    pg2.add((EX["b"], A, BRICK.Temperature_Sensor))
    res = pg2.query("SELECT * WHERE { ?x a brick:Temperature_Sensor }")
    assert len(res) == 1

    pg += pg2
    res = pg.query("SELECT * WHERE { ?x a brick:Temperature_Sensor }")
    assert len(res) == 2



import brickschema
from ontoenv import OntoEnv, Config
from rdflib import Graph, OWL
import pytest
import os
import sys
import io
import pkgutil


def loadGraph(resource) -> brickschema.Graph:
    data = pkgutil.get_data(__name__, resource).decode()
    g = brickschema.Graph()
    g.parse(source=io.StringIO(data), format="turtle")
    return g


# NOTE: Assertions on number of violations is tighly coupled with 1.
# the default and extra (if any) shape files, 2. the data graph.
# Changes in Brick.ttl and pyshacl package, though less frequent
# can have impact, too.


def test_validate_bad(brick_with_imports, shacl_engine):
    dataG = loadGraph("data/badBuilding.ttl")
    # remove imports from the Brick graph
    conforms, _, _ = dataG.validate(
        extra_graphs=[brick_with_imports],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert not conforms, "expect constraint violations in badBuilding.ttl"


def test_validate_ok(brick_with_imports, shacl_engine):
    dataG = loadGraph("data/goodBuilding.ttl")
    conforms, _, report_str = dataG.validate(
        extra_graphs=[brick_with_imports],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert conforms, f"expect no constraint violations in goodBuilding.ttl {report_str}"


def test_useOnlyExtraShapeGraph(shacl_engine):
    dataG = loadGraph("data/badBuilding.ttl")
    shapeG = loadGraph("data/extraShapes.ttl")
    brickG = brickschema.Graph(load_brick=True)
    brickG.remove((None, OWL.imports, None))
    conforms, _, _ = dataG.validate(
        extra_graphs=[shapeG, brickG],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert not conforms, "expect constraint violations in badBuilding.ttl"


def test_useExtraShapeGraph(shacl_engine):
    dataG = loadGraph("data/badBuilding.ttl")
    shapeG = loadGraph("data/extraShapes.ttl")
    brickG = brickschema.Graph(load_brick=True)
    brickG.remove((None, OWL.imports, None))
    conforms, _, _ = dataG.validate(
        extra_graphs=[shapeG, brickG],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert not conforms, "expect constraint violations in badBuilding.ttl"


def test_useExtraOntGraphShapeGraph(shacl_engine):
    dataG = loadGraph("data/badBuilding.ttl")
    ontG1 = loadGraph("data/extraOntology1.ttl")
    ontG2 = loadGraph("data/extraOntology2.ttl")
    brickG = brickschema.Graph(load_brick=True)
    brickG.remove((None, OWL.imports, None))

    # Without extra shapes for the extra ontology files
    # we shouldn't see more violations
    conforms, _, _ = dataG.validate(
        extra_graphs=[ontG1, brickG],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert not conforms, "expect constraint violations in badBuilding.ttl"
    # assert len(result.violationGraphs) == 4, "unexpected # of violations"

    conforms, _, _ = dataG.validate(
        extra_graphs=[ontG1, ontG2, brickG],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert not conforms, "expect constraint violations in badBuilding.ttl"
    # assert len(result.violationGraphs) == 4, "unexpected # of violations"

    shapeG1 = loadGraph("data/extraShapes.ttl")
    shapeG2 = loadGraph("data/extraShapesWithExtraOnt.ttl")

    # Add one extraShape file
    # result = v.validate(dataG, ont_graphs=[ontG1, ontG2], shacl_graphs=[shapeG1])
    conforms, _, _ = dataG.validate(
        extra_graphs=[shapeG1, ontG1, ontG2, brickG],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert not conforms, "expect constraint violations in badBuilding.ttl"
    # assert len(result.violationGraphs) == 9, "unexpected # of violations"

    # Add second extraShape file that goes with the extra ontology
    # result = v.validate(
    #    dataG, ont_graphs=[ontG1, ontG2], shacl_graphs=[shapeG1, shapeG2]
    # )
    conforms, _, _ = dataG.validate(
        extra_graphs=[shapeG1, shapeG2, ontG1, ontG2, brickG],
        engine=shacl_engine,
        min_iterations=1,
        max_iterations=1,
    )
    assert not conforms, "expect constraint violations in badBuilding.ttl"
    # assert len(result.violationGraphs) == 11, "unexpected # of violations"



from datetime import datetime
import logging
from collections import OrderedDict
import uuid
import time
from contextlib import contextmanager
from rdflib import ConjunctiveGraph
from rdflib.graph import BatchAddGraph
from rdflib import plugin, URIRef
from rdflib.store import Store
from rdflib_sqlalchemy import registerplugins
from sqlalchemy import text, Row
import pickle
from .graph import Graph, BrickBase

registerplugins()
logger = logging.getLogger(__name__)

changeset_table_defn = """CREATE TABLE IF NOT EXISTS changesets (
    id TEXT,
    timestamp TIMESTAMP NOT NULL,
    graph TEXT NOT NULL,
    is_insertion BOOLEAN NOT NULL,
    triple BLOB NOT NULL
);"""
changeset_table_idx = """CREATE INDEX IF NOT EXISTS changesets_idx
                        ON changesets (graph, timestamp);"""
redo_table_defn = """CREATE TABLE IF NOT EXISTS redos (
    id TEXT,
    timestamp TIMESTAMP NOT NULL,
    graph TEXT NOT NULL,
    is_insertion BOOLEAN NOT NULL,
    triple BLOB NOT NULL
);"""

_remove_params = ["_delay_init", "brick_version", "load_brick", "load_brick_nightly"]


class PersistentGraph(Graph):
    def __init__(self, uri: str, *args, **kwargs):
        store = plugin.get("SQLAlchemy", Store)(
            identifier="brickschema_persistent_graph"
        )
        kwargs.update({"_delay_init": True})
        super().__init__(store, *args, **kwargs)

        kwargs.update({"create": True})
        for k in _remove_params:
            kwargs.pop(k, None)
        super().open(uri, **kwargs)
        self.uri = uri
        super()._graph_init()


class Changeset(Graph):
    def __init__(self, graph_name):
        super().__init__()
        self.name = URIRef(graph_name)
        self.uid = uuid.uuid4()
        self.additions = []
        self.deletions = []

    def add(self, triple):
        """
        Add a triple to the changeset
        """
        self.additions.append(triple)
        super().add(triple)

    def load_file(self, filename):
        g = Graph()
        g.parse(filename, format="turtle")
        self.additions.extend(g.triples((None, None, None)))
        self += g

        # propagate namespaces
        for pfx, ns in g.namespace_manager.namespaces():
            self.bind(pfx, ns)

    def remove(self, triple):
        self.deletions.append(triple)
        super().remove(triple)


class VersionedGraphCollection(ConjunctiveGraph, BrickBase):
    def __init__(self, uri: str, *args, **kwargs):
        """
        To create an in-memory store, use uri="sqlite://"
        """
        store = plugin.get("SQLAlchemy", Store)(identifier=URIRef("my_store"))
        super().__init__(store, *args, **kwargs)
        self.open(uri, create=True)
        self._precommit_hooks = OrderedDict()
        self._postcommit_hooks = OrderedDict()

        with self.conn() as conn:
            conn.execute(text("PRAGMA journal_mode=WAL;"))
            # conn.execute("PRAGMA synchronous=OFF;")
            conn.execute(text(changeset_table_defn))
            conn.execute(text(changeset_table_idx))
            conn.execute(text(redo_table_defn))

    @property
    def latest_version(self):
        with self.conn() as conn:
            rows = conn.execute(text(
                "SELECT id, timestamp from changesets "
                "ORDER BY timestamp DESC LIMIT 1"
            ))
            res = rows.fetchone()
            return res._asdict() if res else None

    def version_before(self, ts: str) -> str:
        """Returns the version timestamp most immediately
        *before* the given iso8601-formatted timestamp"""
        with self.conn() as conn:
            rows = conn.execute(text(
                "SELECT timestamp from changesets "
                "WHERE timestamp < :ts "
                "ORDER BY timestamp DESC LIMIT 1"),
                                {"ts": ts}
            )
            res = rows.fetchone()
            return res[0]

    def __len__(self):
        # need to override __len__ because the rdflib-sqlalchemy
        # backend doesn't support .count() for recent versions of
        # SQLAlchemy
        return len(list(self.triples((None, None, None))))

    def undo(self):
        """
        Undoes the given changeset. If no changeset is given,
        undoes the most recent changeset.
        """
        if self.latest_version is None:
            raise Exception("No changesets to undo")
        with self.conn() as conn:
            changeset_id = self.latest_version["id"]
            logger.info(f"Undoing changeset {changeset_id}")
            self._graph_at(
                self, conn, self.version_before(self.latest_version["timestamp"])
            )
            conn.execute(
                    text("INSERT INTO redos(id, timestamp, graph, is_insertion, triple) SELECT id, timestamp, graph, is_insertion, triple FROM changesets WHERE id = :id").bindparams(id=changeset_id)
            )
            conn.execute(text("DELETE FROM changesets WHERE id = :id").bindparams(id=changeset_id))

    def redo(self):
        """
        Redoes the most recent changeset.
        """
        with self.conn() as conn:
            redo_record = conn.execute(
                text("SELECT * from redos " "ORDER BY timestamp ASC LIMIT 1")
            ).mappings().fetchone()
            if redo_record is None:
                raise Exception("No changesets to redo")
            changeset_id = redo_record["id"]
            logger.info(f"Redoing changeset {changeset_id}")
            conn.execute(
                    text("INSERT INTO changesets SELECT * FROM redos WHERE id = :id").bindparams(id=changeset_id)
            )
            conn.execute(text("DELETE FROM redos WHERE id = :id").bindparams(id=changeset_id))
            self._graph_at(self, conn, redo_record["timestamp"])
            for row in conn.execute(
                    text("SELECT * from changesets WHERE id = :id").bindparams(id=changeset_id)
            ).mappings():
                triple = pickle.loads(row["triple"])
                graph = self.get_context(redo_record["graph"])
                if row["is_insertion"]:
                    graph.remove((triple[0], triple[1], triple[2]))
                else:
                    graph.add((triple[0], triple[1], triple[2]))

    def versions(self, graph=None):
        """
        Return a list of all versions of the provided graph; defaults
        to the union of all graphs
        """
        with self.conn() as conn:
            if graph is None:
                rows = conn.execute(text(
                    "SELECT DISTINCT id, graph, timestamp from changesets "
                    "ORDER BY timestamp DESC"
                    ))
            else:
                rows = conn.execute(text(
                    "SELECT DISTINCT id, graph, timestamp from changesets "
                    "WHERE graph = :g ORDER BY timestamp DESC").bindparams(g=graph)
                )
            return list(rows)

    def add_precommit_hook(self, hook):
        self._precommit_hooks[hook.__name__] = hook

    def add_postcommit_hook(self, hook):
        self._postcommit_hooks[hook.__name__] = hook

    @property
    def conn(self):
        return self.store.engine.begin

    @contextmanager
    def new_changeset(self, graph_name, ts=None):
        if not isinstance(graph_name, URIRef):
            graph_name = URIRef(graph_name)
        namespaces = []
        buffered_adds = []
        buffered_removes = []
        with self.conn() as conn:
            transaction_start = time.time()
            cs = Changeset(graph_name)
            yield cs
            if ts is None:
                ts = datetime.now().isoformat()
            # delta by the user. We need to invert the changes so that they are expressed as a "backward"
            # delta. This means that we save the deletions in the changeset as "inserts", and the additions
            # as "deletions".
            if cs.deletions:
                for triple in cs.deletions:
                    conn.execute(
                        text("INSERT INTO changesets VALUES (:uid, :ts, :graph, :deletion, :triple)").bindparams(
                            uid=str(cs.uid),
                            ts=ts,
                            graph=str(graph_name),
                            deletion=True,
                            triple=pickle.dumps(triple),
                        )
                    )
                for triple in cs.deletions:
                    buffered_removes.append(triple)
                #graph = self.get_context(graph_name)
                #for triple in cs.deletions:
                #    graph.remove(triple)
            if cs.additions:
                for triple in cs.additions:
                    conn.execute(
                        text("INSERT INTO changesets VALUES (:uid, :ts, :graph, :deletion, :triple)").bindparams(
                            uid=str(cs.uid),
                            ts=ts,
                            graph=str(graph_name),
                            deletion=False,
                            triple=pickle.dumps(triple),
                        )
                    )
                for triple in cs.additions:
                    buffered_adds.append(triple)
                # with BatchAddGraph(
                #     self.get_context(graph_name), batch_size=10000
                # ) as graph:
                #     for triple in cs.additions:
                #         graph.add(triple)

            # take care of precommit hooks
            transaction_end = time.time()
            for hook in self._precommit_hooks.values():
                hook(self)
            # keep track of namespaces so we can add them to the graph
            # after the commit
            namespaces.extend(cs.namespace_manager.namespaces())

            # # finally, remove all of the 'redos'
            # conn.execute("DELETE FROM redos")
            # # and remove all of the 'changesets' that come after us
            logging.info(
                f"Committing after {transaction_end - transaction_start} seconds"
            )
        # add the buffered changes to the graph
        print([(type(c.identifier), c.identifier) for c in self.contexts()])
        graph = self.get_context(graph_name)
        for triple in buffered_removes:
            print(f"Removing {triple}")
            graph.remove(triple)
        with BatchAddGraph(graph, batch_size=10000) as graph:
            for triple in buffered_adds:
                print(f"Adding {triple}")
                graph.add(triple)
        print(f"Self graph has {len(self)} triples")
        # loop through all of the contexts and print length
        # update namespaces
        for pfx, ns in namespaces:
            self.bind(pfx, ns)
        for hook in self._postcommit_hooks.values():
            hook(self)
        self._latest_version = ts
        for c in self.contexts():
            print(f"{c.identifier} has {len(c)} triples")

    def latest(self, graph):
        return self.get_context(graph)

    def graph_at(self, timestamp=None, graph=None):
        # setup graph and bind namespaces
        g = Graph()
        for pfx, ns in self.namespace_manager.namespaces():
            g.bind(pfx, ns)
        # if graph is specified, only copy triples from that graph.
        # otherwise, copy triples from all graphs.
        if graph is not None:
            for t in self.get_context(graph).triples((None, None, None)):
                g.add(t)
        else:
            # TODO: this doesn't work for some reason
            for t in self.triples((None, None, None)):
                g.add(t)
        with self.conn() as conn:
            return self._graph_at(g, conn, timestamp, graph)

    def _graph_at(self, alter_graph, conn, timestamp=None, graph=None):
        """
        Return *copy* of the graph at the given timestamp. Chooses the most recent timestamp
        that is less than or equal to the given timestamp.
        """
        if timestamp is None:
            timestamp = datetime.now().isoformat()
        if isinstance(timestamp, (dict, Row)):
            timestamp = timestamp["timestamp"]

        print(f"Getting graph {graph} ({type(graph)}) at {timestamp}", type(timestamp))
        # print # of rows in changesets
        print(f"Changesets has {len(list(conn.execute(text('SELECT * FROM changesets'))))} rows")
        if graph is not None:
            rows = conn.execute(
                    text("SELECT * FROM changesets WHERE graph = :g AND timestamp > :ts ORDER BY timestamp DESC").bindparams(
                        g=graph, ts=timestamp
                    )
            )
        else:
            rows = conn.execute(
                    text("SELECT * FROM changesets WHERE timestamp > :ts ORDER BY timestamp DESC").bindparams(
                        ts=timestamp
                    )
            )
        for row in rows.mappings():
            print(f"Row: {row}")
            triple = pickle.loads(row["triple"])
            if row["is_insertion"]:
                print(f"Adding {triple}")
                alter_graph.add((triple[0], triple[1], triple[2]))
            else:
                print(f"Removing {triple}")
                alter_graph.remove((triple[0], triple[1], triple[2]))
        return alter_graph



import rdflib
from brickschema.inference import HaystackInferenceSession
from brickschema.namespaces import BRICK, A
from rdflib import Namespace, Graph
from typer import progressbar


class HaystackRDFInferenceSession(HaystackInferenceSession):
    def __init__(self, namespace):
        super().__init__(namespace)

    def infer_model(self, graph: rdflib.Graph):
        query = """
        SELECT DISTINCT ?instance (GROUP_CONCAT(?markers; SEPARATOR=" ") AS ?p) WHERE {
          ?instance ph:hasTag ?markers . ?markers (a|ph:is|rdfs:subClassOf)* ph:marker .
        } GROUP BY ?instance
        """
        bg = Graph()
        bg.bind("brick", BRICK)
        bg.bind("hs", Namespace("https://project-haystack.dev/example#"))
        with progressbar(graph.query(query)) as results:
            for instance, markers in results:
                marker_tags = [
                    marker.split("#")[-1] for marker in str(markers).split(" ")
                ]
                marker_tags = list(set(marker_tags))
                for f in self._filters:
                    marker_tags = list(filter(f, marker_tags))
                # translate tags
                entity_tagset = list(self._translate_tags(marker_tags))
                # infer tags for single entity
                triples, _ = self.infer_entity(entity_tagset, identifier="id")
                for triple in triples:
                    if triple[1] == A:
                        graph.add((instance, A, triple[2]))
                        graph.add((instance, BRICK.label, graph.label(instance)))



from brickschema import Graph, GraphCollection
from brickschema.namespaces import BRICK, UNIT, A
from rdflib import Namespace, Literal, URIRef


def test_specific_classes():
    g = Graph(load_brick=True)

    classlist = [BRICK.HVAC_Equipment, BRICK.Equipment, BRICK.Fan, BRICK.Discharge_Fan]
    specific = g.get_most_specific_class(classlist)
    assert specific == [BRICK.Discharge_Fan]

    classlist = [
        BRICK.HVAC_Equipment,
        BRICK.Equipment,
        BRICK.Fan,
        BRICK.Discharge_Fan,
        BRICK.Exhaust_Fan,
    ]
    specific = g.get_most_specific_class(classlist)
    assert specific == [BRICK.Discharge_Fan, BRICK.Exhaust_Fan]

    classlist = [BRICK.HVAC_Equipment, BRICK.Fan]
    specific = g.get_most_specific_class(classlist)
    assert specific == [BRICK.Fan]

    classlist = [BRICK.HVAC_Equipment, BRICK.Chiller, BRICK.Absorption_Chiller]
    specific = g.get_most_specific_class(classlist)
    assert specific == [BRICK.Absorption_Chiller]


def test_add_fancy():
    g = Graph()

    EX = Namespace("urn:ex#")
    g.bind("ex", EX)

    g.add(
        (EX.A, BRICK.area, [(BRICK.value, Literal(100)), (BRICK.hasUnit, UNIT["M3"])])
    )

    res = list(
        g.query(
            """SELECT ?area ?unit WHERE {
                            ?x brick:area/brick:value ?area .
                            ?x brick:area/brick:hasUnit ?unit
                        }"""
        )
    )
    assert len(res) == 1


def test_operator_overload():
    EX = Namespace("urn:ex#")

    g1 = Graph()
    g1.add((EX["a"], A, BRICK["Sensor"]))

    g2 = Graph()
    g2.add((EX["b"], A, BRICK["Sensor"]))

    g = g1 + g2

    g.expand("owlrl")

    res = g.query("SELECT * WHERE { ?x a brick:Sensor }")
    assert len(res) == 2, "Should have 2 sensors from adding graphs"


def test_collection():
    g = GraphCollection(load_brick=True)
    assert URIRef(BRICK) in g.graph_names
    assert len(g.graph_names) == 2

    EX1 = Namespace("urn:ex1#")
    bldg = g.graph(EX1)
    bldg.add((EX1["a"], A, BRICK["Sensor"]))

    assert len(g.graph_names) == 3
    assert URIRef(EX1) in g.graph_names
    assert URIRef(BRICK) in g.graph_names
    res = g.query("SELECT * WHERE { ?x a brick:Sensor }")
    assert len(res) == 1, "Should have 1 sensor from adding graph"

    EX2 = Namespace("urn:ex2#")
    bldg = g.graph(EX2)
    bldg.add((EX2["a"], A, BRICK["Sensor"]))

    assert len(g.graph_names) == 4
    assert URIRef(EX1) in g.graph_names
    assert URIRef(EX2) in g.graph_names
    assert URIRef(BRICK) in g.graph_names
    res = g.query("SELECT * WHERE { ?x a brick:Sensor }")
    assert len(res) == 2, "Should have 2 sensors from adding graph"

    # This needs more work!
    # # option 1: GraphCollection is the 'database' of multiple buildings
    # # - need to be able to pull a view of single building + Brick + other supporting ontologies
    # # - one copy of Brick
    # # - TODO: where do the 'reasoned' triples go? Can we direct some reasoned triples to the "original" graph?
    # # - TODO: need to have a r/w view of the graph
    # # - TODO: can we have the store support querying a subset of the graphs in a single Dataset/ConjunctiveGraph? RDFlib may be interested in this
    # # option 2: GraphCollection is 'database' of one building
    # # - no need to query a subset of the graphs
    # sg = g.subset_with([EX2, BRICK])
    # assert len(sg.graph_names) == 3, sg.graph_names
    # assert URIRef(EX1) not in sg.graph_names
    # assert URIRef(EX2) in sg.graph_names
    # assert URIRef(BRICK) in sg.graph_names
    # sg_ex2 = sg.graph(EX2)
    # sg_ex2.add((EX2["b"], A, BRICK["Sensor"]))
    # res = sg.query("SELECT * WHERE { ?x a brick:Sensor }")
    # assert len(res) == 2, "Should have 2 sensors in remaining graph (we added one)"

    # assert len(g.graph_names) == 4
    # assert URIRef(EX1) in g.graph_names
    # assert URIRef(EX2) in g.graph_names
    # assert URIRef(BRICK) in g.graph_names

    # res = g.query("SELECT * WHERE { ?x a brick:Sensor }")
    # assert len(res) == 3, "Should now have 3 sensors from adding graph"



from brickschema.persistent import VersionedGraphCollection
from brickschema.namespaces import BRICK, UNIT, A
from rdflib import Namespace, Literal, URIRef


def test_versioned_graph():
    g = VersionedGraphCollection("sqlite://")
    EX = Namespace("urn:ex#")
    g.bind("ex", EX)
    assert len(g) == 0
    with g.new_changeset("abc") as cs:
        cs.add((EX["a"], A, BRICK.Sensor))
    assert len(g) == 1
    with g.new_changeset("abc") as cs:
        cs.add((EX["b"], A, BRICK.Sensor))
    assert len(g) == 2
    g.undo()
    assert len(g) == 1
    g.redo()
    assert len(g) == 2


def test_versioned_graph_at():
    g = VersionedGraphCollection("sqlite://")
    EX = Namespace("urn:ex#")
    g.bind("ex", EX)
    assert len(g) == 0
    with g.new_changeset("abc") as cs:
        cs.add((EX["a"], A, BRICK.Sensor))
    assert len(g) == 1
    _tmp = g.graph_at(g.latest_version["timestamp"])
    assert len(_tmp) == 1
    with g.new_changeset("abc") as cs:
        cs.add((EX["b"], A, BRICK.Sensor))
    _tmp = g.graph_at(g.latest_version["timestamp"])
    assert len(_tmp) == 2

    assert len(g) == 2
    g.undo()
    assert len(g) == 1
    g.redo()
    assert len(g) == 2



#!/usr/bin/env python

import brickschema
import shutil
import sys

if len(sys.argv) > 1:
    brick_file = sys.argv[1]
else:
    brick_file = None

sess = brickschema.inference.TagInferenceSession(
    rebuild_tag_lookup=True, brick_file=brick_file
)
shutil.copyfile("taglookup.pickle", "brickschema/ontologies/1.3/taglookup.pickle")



import re
from pathlib import Path
from typing import Optional, List

import importlib_resources
import rdflib
import typer
from typer import progressbar

from brickschema.brickify.util import bind_namespaces, load_config


class Handler:
    def __init__(
        self,
        source: Optional[str] = "input.ttl",
        input_format: Optional[str] = "turtle",
        module_path: Optional[List[str]] = None,
        config_file: Optional[Path] = None,
    ):
        """
        Handler class responsible for performing end to end conversion
        (including ingestion, translation, and clean up).

        :param source: A filepath/URL
        :param input_format: Input format of the file
        :param module_path: Path to default template files in the package ([<dot-separate-module-path>, <template-filename>])
        :param config_file: Custom conversion configuration file
        """
        self.graph = rdflib.Graph()
        self.source = source
        self.input_format = input_format
        if config_file:
            with open(config_file, "r") as config:
                self.config = load_config(config, config_file)
        elif module_path:
            config_file = module_path[-1]
            with importlib_resources.path(*module_path) as data_file:
                with open(data_file, "r") as config:
                    self.config = load_config(config, config_file)
        else:
            typer.echo(
                typer.style(
                    "[ERROR] No configuration specified!",
                    fg=typer.colors.RED,
                )
            )

    def update_namespaces(
        self,
        building_prefix=None,
        building_namespace=None,
        site_prefix=None,
        site_namespace=None,
    ):
        """
        Updates prefixes and namespaces from the templates for the site and the building before conversion.

        :param building_prefix: Building prefix (default: bldg)
        :param building_namespace: Building namespace (default: https://example.com/bldg#)
        :param site_prefix: Site prefix (default: site)
        :param site_namespace: Site namespace (default: https://example.com/site#)
        """
        self.config["namespace_prefixes"][building_prefix] = building_namespace
        self.config["namespace_prefixes"][site_prefix] = site_namespace
        bind_namespaces(self.graph, self.config["namespace_prefixes"])
        for operation in self.config["operations"]:
            if "query" in operation:
                operation["query"] = re.sub(
                    "bldg:", f"{building_prefix}:", operation["query"]
                )
                operation["query"] = re.sub(
                    "site:", f"{site_prefix}:", operation["query"]
                )
            if "data" in operation:
                operation["data"] = re.sub(
                    "bldg:", f"{building_prefix}:", operation["data"]
                )
                operation["data"] = re.sub(
                    "site:", f"{site_prefix}:", operation["data"]
                )
            if "template" in operation:
                operation["template"] = re.sub(
                    "bldg:", f"{building_prefix}:", operation["template"]
                )
                operation["template"] = re.sub(
                    "site:", f"{site_prefix}:", operation["template"]
                )

    def ingest_data(self):
        """
        Ingests the data from files to the memory. The default option
        for the base handler is to parse a source graph in the specified input format
        to self.graph.
        """
        self.graph.parse(self.source, format=rdflib.util.guess_format(self.source))

    def translate(self):
        """
        Performs SPARQL based operations sequentially over the output graph.
        """
        if not self.config["operations"]:
            return
        with progressbar(self.config["operations"]) as operations:
            for operation in operations:
                if "data" in operation:
                    query = f"INSERT DATA {{{{ {operation['data']} }}}}"
                elif "query" in operation:
                    query = operation["query"]
                if not query:
                    continue
                query = query.format_map({})
                try:
                    self.graph.update(query)
                except Exception as e:
                    print(e)
                    print(query)

    def infer(self):
        """
        In the implementations of specific handlers, this method would be overridden to perform and append data to the output graph
        based on additional inference.
        """
        pass

    def clean_up(self):
        """
        In the implementations of specific handlers, this method would be overridden to clean the graphs, remove intermediate data, etc.
        """
        pass

    def convert(self, building_prefix, building_namespace, site_prefix, site_namespace):
        """
        Performs the conversion based on the base conversion sequence. By default, it updates the namespaces,
        ingests the data, translates the data to brick primarily using SPARQL based operations, performs and adds data
        from additional inferences, cleans up the output graph.


        :param building_prefix: Building prefix (default: bldg)
        :param building_namespace: Building namespace (default: https://example.com/bldg#)
        :param site_prefix: Site prefix (default: site)
        :param site_namespace: Site namespace (default: https://example.com/site#)
        :returns: A Brick graph (rdflib.Graph)
        """
        self.update_namespaces(
            building_prefix, building_namespace, site_prefix, site_namespace
        )
        self.ingest_data()
        self.translate()
        self.infer()
        self.clean_up()
        return self.graph



from rdflib import Graph
from typer.testing import CliRunner

from brickschema.brickify.main import app
from brickschema.brickify.main import convert
from brickschema.namespaces import BRICK

runner = CliRunner()


def test_rac():
    result = runner.invoke(
        app, ["tests/data/brickify/RAC/rac.xls", "--input-type", "rac"], input="\n"
    )
    print(result.stdout)
    assert result.exit_code == 0


def test_haystack_ttl():
    result = runner.invoke(
        app,
        [
            "https://project-haystack.org/example/download/charlie.ttl",
            "--input-type",
            "haystack-v4",
            "--output",
            "tests/data/brickify/haystack-v4/charlie.ttl.brick.ttl",
        ],
    )
    assert result.exit_code == 0, result.stdout


def test_jinja2():
    result = runner.invoke(
        app,
        [
            "tests/data/brickify/jinja2/sheet.csv",
            "--input-type",
            "csv",
            "--config",
            "tests/data/brickify/jinja2/template.yml",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0


def test_rdf():
    result = runner.invoke(
        app,
        [
            "tests/data/brickify/rdf/input.ttl",
            "--input-type",
            "rdf",
            "--config",
            "tests/data/brickify/rdf/template.yml",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0


def test_tsv():
    result = runner.invoke(
        app,
        [
            "tests/data/brickify/tsv/sheet.tsv",
            "--input-type",
            "tsv",
            "--config",
            "tests/data/brickify/tsv/template.yml",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0



"""
The `namespaces` module provides pointers to standard Brick namespaces
and related ontology namespaces
wrapper class and convenience methods for a Brick graph
"""
from rdflib import Namespace

BRICK11 = Namespace("https://brickschema.org/schema/1.1/Brick#")
TAG11 = Namespace("https://brickschema.org/schema/1.1/BrickTag#")
BSH11 = Namespace("https://brickschema.org/schema/1.1/BrickShape#")

# all versions of Brick > 1.1 have these namespaces
BRICK = Namespace("https://brickschema.org/schema/Brick#")
TAG = Namespace("https://brickschema.org/schema/BrickTag#")
BSH = Namespace("https://brickschema.org/schema/BrickShape#")
REF = Namespace("https://brickschema.org/schema/Brick/ref#")

# other ontologies
BACNET = Namespace("http://data.ashrae.org/bacnet/2020#")

# defaults
OWL = Namespace("http://www.w3.org/2002/07/owl#")
RDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
SH = Namespace("http://www.w3.org/ns/shacl#")

# QUDT namespaces
QUDT = Namespace("http://qudt.org/schema/qudt/")
QUDTQK = Namespace("http://qudt.org/vocab/quantitykind/")
QUDTDV = Namespace("http://qudt.org/vocab/dimensionvector/")
UNIT = Namespace("http://qudt.org/vocab/unit/")

A = RDF.type


def bind_prefixes(graph, brick_version="1.3"):
    """
    Associate common prefixes with the graph
    """
    graph.bind("rdf", RDF)
    graph.bind("owl", OWL)
    graph.bind("rdfs", RDFS)
    graph.bind("skos", SKOS)
    graph.bind("sh", SH)
    graph.bind("qudtqk", QUDTQK)
    graph.bind("qudt", QUDT)
    graph.bind("unit", UNIT)
    graph.bind("ref", REF)
    graph.bind("bacnet", BACNET)

    if brick_version == "1.1":
        graph.bind("brick", BRICK11)
        graph.bind("tag", TAG11)
        graph.bind("bsh", BSH11)
    else:
        graph.bind("brick", BRICK)
        graph.bind("tag", TAG)
        graph.bind("bsh", BSH)



import subprocess
import platform
import tempfile
import rdflib
from rdflib import OWL, SH
from rdflib.term import BNode, URIRef, _SKOLEM_DEFAULT_AUTHORITY, rdflib_skolem_genid
from pathlib import Path
from typing import Optional
from urllib.parse import urljoin


# monkeypatch BNode.skolemize with a new function
def _new_bnode_skolemize(
    self, authority: Optional[str] = None, basepath: Optional[str] = None
) -> URIRef:
    """Create a URIRef "skolem" representation of the BNode, in accordance
    with http://www.w3.org/TR/rdf11-concepts/#section-skolemization

    .. versionadded:: 4.0
    """
    if authority is None:
        authority = _SKOLEM_DEFAULT_AUTHORITY
    if basepath is None:
        basepath = rdflib_skolem_genid
    skolem = "%s%s" % (basepath, str(self).replace(" ", "_"))
    return URIRef(urljoin(authority, skolem))


BNode.skolemize = _new_bnode_skolemize
_MAX_EXTERNAL_LOOPS = 3


def infer(
    data_graph: rdflib.Graph, ontologies: rdflib.Graph, max_iterations: int = 100
):
    # remove imports
    imports = data_graph.triples((None, OWL.imports, None))
    data_graph.remove((None, OWL.imports, None))

    # skolemize before inference
    data_graph_skolemized = data_graph.skolemize()

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Define the target path within the temporary directory
        target_file_path = temp_dir_path / "data.ttl"
        (data_graph_skolemized + ontologies).serialize(
            target_file_path, format="turtle"
        )

        # set the SHACL_HOME environment variable to point to the shacl-1.4.2 directory
        # so that the shaclinfer.sh script can find the shacl.jar file
        env = {"SHACL_HOME": str(Path(__file__).parent / "topquadrant_shacl")}
        # get the shacl-1.4.2/bin/shaclinfer.sh script from brickschema.bin in this package
        # using pkgutil. If using *nix, use .sh; else if on windows use .bat
        if platform.system() == "Windows":
            script = [
                str(Path(__file__).parent / "topquadrant_shacl/bin/shaclinfer.bat")
            ]
        else:
            script = [
                "/bin/sh",
                str(Path(__file__).parent / "topquadrant_shacl/bin/shaclinfer.sh"),
            ]

        # Initialize the size of the graph
        previous_size = 0
        current_size = len(data_graph_skolemized)
        current_iter = 0

        # Run the shaclinfer multiple times until the skolemized data graph stops changing in size
        while previous_size != current_size and current_iter < _MAX_EXTERNAL_LOOPS:
            (data_graph_skolemized + ontologies).serialize(
                target_file_path, format="turtle"
            )
            try:
                print(f"Running {script} -datafile {target_file_path}")
                output = subprocess.check_output(
                    [
                        *script,
                        "-datafile",
                        target_file_path,
                        "-maxiterations",
                        str(max_iterations),
                    ],
                    stderr=subprocess.STDOUT,
                    universal_newlines=True,
                    env=env,
                )
            except subprocess.CalledProcessError as e:
                output = e.output  # Capture the output of the failed subprocess
            # Write logs to a file in the temporary directory (or the desired location)
            inferred_file_path = temp_dir_path / "inferred.ttl"
            with open(inferred_file_path, "w") as f:
                for line in output.splitlines():
                    if "::" not in line:
                        f.write(f"{line}\n")
            inferred_triples = rdflib.Graph()
            inferred_triples.parse(inferred_file_path, format="turtle")
            print(f"Got {len(inferred_triples)} inferred triples")
            for s, p, o in inferred_triples:
                if isinstance(s, BNode) or isinstance(o, BNode):
                    continue
                data_graph_skolemized.add((s, p, o))

            # Update the size of the graph
            previous_size = current_size
            current_size = len(data_graph_skolemized)
            current_iter += 1

        expanded_graph = data_graph_skolemized.de_skolemize()
        # add imports back in
        for imp in imports:
            expanded_graph.add(imp)
        return expanded_graph


def validate(data_graph: rdflib.Graph, shape_graphs: rdflib.Graph):
    # remove imports
    data_graph.remove((None, OWL.imports, None))

    # set the SHACL_HOME environment variable to point to the shacl-1.4.2 directory
    # so that the shaclinfer.sh script can find the shacl.jar file
    env = {"SHACL_HOME": str(Path(__file__).parent / "topquadrant_shacl")}
    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Define the target path within the temporary directory
        target_file_path = temp_dir_path / "data.ttl"

        inferred_graph = infer(data_graph, shape_graphs)

        inferred_graph.serialize(target_file_path, format="ttl")

        # get the shacl-1.4.2/bin/shaclvalidate.sh script from the same directory
        # as this file
        if platform.system() == "Windows":
            script = [
                str(Path(__file__).parent / "topquadrant_shacl/bin/shaclvalidate.bat")
            ]
        else:
            script = [
                "/bin/sh",
                str(Path(__file__).parent / "topquadrant_shacl/bin/shaclvalidate.sh"),
            ]
        try:
            print(f"Running {script} -datafile {target_file_path}")
            output = subprocess.check_output(
                [*script, "-datafile", target_file_path],
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                env=env,
            )
        except subprocess.CalledProcessError as e:
            output = e.output  # Capture the output of the failed subprocess

        # Write logs to a file in the temporary directory (or the desired location)
        report_file_path = temp_dir_path / "report.ttl"
        with open(report_file_path, "w") as f:
            for line in output.splitlines():
                if "::" not in line:  # filter out log output
                    f.write(f"{line}\n")

        report_g = rdflib.Graph()
        report_g.parse(report_file_path, format="turtle")

        # check if there are any sh:resultSeverity sh:Violation predicate/object pairs
        has_violation = len(
            list(report_g.subjects(predicate=SH.resultSeverity, object=SH.Violation))
        )
        conforms = len(
            list(report_g.subjects(predicate=SH.conforms, object=rdflib.Literal(True)))
        )
        validates = not has_violation or conforms

        return validates, report_g, str(report_g.serialize(format="turtle"))



"""
The `merge` module implements data integration methods for merging Brick graphs
together. This is based on techniques described in 'Shepherding Metadata Through
the Building Lifecycle' published in BuildSys 2020
"""
from colorama import init as colorama_init
from colorama import Fore, Style

from rdflib import URIRef
from collections import defaultdict
import dedupe
from .graph import Graph
from dedupe._typing import (
    TrainingData,
    Literal,
)
import sys
from dedupe.core import unique
from typing import List, Tuple, Any

colorama_init()
DEBUG = False


def _unpack_linked_records(linked_records):
    s = set()
    for (e1, e2), _ in linked_records:
        s.add(e1)
        s.add(e2)
    return s


def unify_entities(G, e1, e2):
    """
    Replaces all instances of e2 with e1 in graph G
    """
    print(Style.BRIGHT + Fore.CYAN + f"Unifying {e1} and {e2}" + Style.RESET_ALL)
    e1 = URIRef(e1)
    e2 = URIRef(e2)
    pos = G.predicate_objects(subject=e2)
    for (p, o) in pos:
        G.remove((e2, p, o))
        G.add((e1, p, o))

    sps = G.subject_predicates(object=e2)
    for (s, p) in sps:
        G.remove((s, p, e2))
        G.add((s, p, e1))


def get_entity_feature_vectors(g, namespace):
    """
    Returns a dictionary of features for each entity in graph 'g'.

    Entities are any node with at least one `rdf:type` edge that is in the given namespace.
    """
    entities = g.query(
        f"""SELECT ?ent ?type ?label WHERE {{
        ?ent rdf:type ?type .
        OPTIONAL {{ ?ent rdfs:label ?label }} .
        FILTER(STRSTARTS(STR(?ent), STR(<{namespace}>)))
    }}"""
    )
    features = defaultdict(lambda: defaultdict(set))
    for row in entities:
        entity, etype, label = row
        features[str(entity)]["type"].add(str(etype))
        features[str(entity)]["label"].add(str(label))
        features[str(entity)]["uri"].add(str(entity))

    if DEBUG:
        for ent, featurelist in features.items():
            print(ent)
            for name, vals in featurelist.items():
                print(f"   {name} = {vals}")

    return features


def flatten_features(features):
    for _, flist in features.items():
        for k, v in flist.items():
            if isinstance(v, (list, set)):
                flist[k] = list(v)[0]


def cluster_by_type(g1, g2, namespace):
    clusters = defaultdict(lambda: {"g1": {}, "g2": {}})
    g1_features = get_entity_feature_vectors(g1, namespace)
    g2_features = get_entity_feature_vectors(g2, namespace)
    for g1_ent, g1_ent_feat in g1_features.items():
        for etype in g1_ent_feat["type"]:
            flat_features = g1_ent_feat.copy()
            flat_features["type"] = etype
            clusters[etype]["g1"][g1_ent] = flat_features

    for g2_ent, g2_ent_feat in g2_features.items():
        for etype in g2_ent_feat["type"]:
            flat_features = g2_ent_feat.copy()
            flat_features["type"] = etype
            clusters[etype]["g2"][g2_ent] = flat_features

    if DEBUG:
        for ent_type, cluster in clusters.items():
            print(ent_type)
            print("   ", cluster["g1"].keys())
            print("   ", cluster["g2"].keys())
            print()

    return clusters


def _merge_features(fields, g1_features, g2_features):
    while True:
        if len(g1_features) == len(g2_features) == 1:
            linked_records = (
                (list(g1_features.keys())[0], list(g2_features.keys())[0]),
                1.0,
            )
            g1_features = []
            g2_features = []
            break

        linker = dedupe.RecordLink(fields)
        linker.prepare_training(g1_features, g2_features)
        confirmed_matches = console_label(linker)
        linker.train()
        linked_records = linker.join(
            g1_features, g2_features, 0.0, constraint="one-to-one"
        )

        # remove records from linked_records that are in confirmed_matches
        for (e1, e2) in confirmed_matches:
            idx = 0
            while idx < len(linked_records):
                pair = linked_records[idx]
                for (pair, _) in linked_records:
                    if e1 in pair or e2 in pair:
                        linked_records.pop(idx)
                        break
                idx += 1

        # replace linked record scores with 1.0 if the user explicitly
        # marked them as equivalent. Then fill in other user-marked pairs
        # of linked records at the end
        idx = 0
        while idx < len(confirmed_matches):
            pair = confirmed_matches[idx]
            for lidx, (lpair, _) in enumerate(linked_records):
                if pair == lpair:
                    linked_records[lidx] = (pair, 1.0)
                    confirmed_matches.pop(idx)
                    idx -= 1  # cancel out the increment
                    break
            idx += 1
        linked_records.extend([(pair, 1.0) for pair in confirmed_matches])

        print(
            Style.BRIGHT + Fore.YELLOW + "Is this matching correct?" + Style.RESET_ALL
        )
        for (e1, e2), similarity in linked_records:
            for field in unique(field["field"] for field in fields):
                g1_val = g1_features[e1][field]
                g2_val = g2_features[e2][field]
                print(f"{g1_val:<50} | {g2_val:<50}")
            print(f"Similarity: {similarity}")
            print("-" * 20)
        ans = input(Fore.YELLOW + "[y/n]? " + Style.RESET_ALL)
        if ans.lower() == "y":
            print(
                Fore.GREEN
                + "All correct! Moving on to any stragglers"
                + Style.RESET_ALL
            )
            break
        else:
            print(Fore.RED + "Re-labeling..." + Style.RESET_ALL)
    linked_entities = _unpack_linked_records(linked_records)
    if len(linked_entities) != len(g1_features) or len(linked_entities) != len(
        g2_features
    ):
        leftover_g1 = set(g1_features.keys()).difference(linked_entities)
        leftover_g2 = set(g2_features.keys()).difference(linked_entities)
        leftover_g1 = {k: v for (k, v) in g1_features.items() if k in leftover_g1}
        leftover_g2 = {k: v for (k, v) in g2_features.items() if k in leftover_g2}
    return linked_records, leftover_g1, leftover_g2


def merge_type_cluster(g1, g2, namespace, similarity_threshold=0.9, merge_types=None):
    merge_types = list(map(str, get_common_types(g1, g2, namespace)))
    _g1 = Graph().load_file("Brick.ttl").from_triples(g1.triples((None, None, None)))
    _g1.expand("brick")

    _g2 = Graph().load_file("Brick.ttl").from_triples(g2.triples((None, None, None)))
    _g2.expand("brick")
    clusters = cluster_by_type(_g1, _g2, namespace)

    G = g1 + g2

    linked = set()
    for etype, cluster in clusters.items():
        if merge_types and etype not in merge_types:
            continue
        print(
            Style.BRIGHT
            + Fore.YELLOW
            + f"Handling clusters for {etype}"
            + Style.RESET_ALL
        )
        # if not same # of entities in both clusters,
        # then type alignment will be less successful
        for e in linked:
            if e in cluster["g1"]:
                cluster["g1"].pop(e)
            if e in cluster["g2"]:
                cluster["g2"].pop(e)
        if not len(cluster["g1"]) or not len(cluster["g2"]):
            continue
        g1_features = cluster["g1"].copy()
        g2_features = cluster["g2"].copy()
        flatten_features(g1_features)
        flatten_features(g2_features)
        fields = [
            {"field": "uri", "type": "String"},
            # {"field": "type", "type": "String"},
            {"field": "label", "type": "String"},
        ]
        while True:
            linked_records, leftover_g1, leftover_g2 = _merge_features(
                fields, g1_features, g2_features
            )
            for link in linked_records:
                (e1, e2), similarity = link
                if similarity < similarity_threshold:
                    print(
                        Fore.RED
                        + f"cannot merge {e1}, {e2} due to similarity threshold {similarity} < {similarity_threshold}"
                        + Style.RESET_ALL
                    )
                    continue
                linked.add(e1)
                linked.add(e2)
                unify_entities(G, e1, e2)
            if leftover_g1 and leftover_g2 and len(leftover_g1) and len(leftover_g2):
                print(
                    Style.BRIGHT
                    + Fore.YELLOW
                    + "More entities left to merge"
                    + Style.RESET_ALL
                )
                g1_features = leftover_g1
                g2_features = leftover_g2
                continue
            break
    return G


# def merge_record_linkage(g1, g2, namespace):
#     g1_features = get_entity_feature_vectors(g1, namespace)
#     g2_features = get_entity_feature_vectors(g2, namespace)
#
#     flatten_features(g1_features)
#     flatten_features(g2_features)
#
#     fields = [
#         {"field": "uri", "type": "String"},
#         {"field": "type", "type": "String"},
#         {"field": "label", "type": "String"},
#     ]
#     linker = dedupe.RecordLink(fields)
#
#     linker.prepare_training(g2_features, g1_features)
#     dedupe.console_label(linker)
#     linker.train()
#     linked_records = linker.join(g1_features, g2_features, 0.0, constraint="one-to-one")


def get_common_types(g1, g2, namespace):
    """
    Returns the list of types that are common to both graphs. A type is included
    if both graphs have instances of that type
    """
    g1ents = g1.query(
        f"""SELECT DISTINCT ?type WHERE {{
        ?ent rdf:type ?type .
        FILTER(STRSTARTS(STR(?ent), STR(<{namespace}>)))
    }}"""
    )

    g2ents = g2.query(
        f"""SELECT DISTINCT ?type WHERE {{
        ?ent rdf:type ?type .
        FILTER(STRSTARTS(STR(?ent), STR(<{namespace}>)))
    }}"""
    )

    g1types = set([x[0] for x in g1ents])
    g2types = set([x[0] for x in g2ents])
    return list(g1types.intersection(g2types))


# using code from dedupe.io
def console_label(deduper: dedupe.api.ActiveMatching) -> None:  # noqa: C901
    """
    Train a matcher instance (Dedupe, RecordLink, or Gazetteer) from the command line.
    Example

    .. code:: python

       > deduper = dedupe.Dedupe(variables)
       > deduper.prepare_training(data)
       > dedupe.console_label(deduper)
    """

    confirmed_matches = []

    finished = False
    use_previous = False
    fields = unique(field.field for field in deduper.data_model.primary_fields)

    buffer_len = 1  # Max number of previous operations
    examples_buffer: List[Tuple[Any, Literal["match", "distinct", "uncertain"]]] = []
    uncertain_pairs: List[Any] = []

    # don't re-use items that are confirmed with a mapping
    mapped_items = set()

    while not finished:
        if use_previous:
            record_pair, _ = examples_buffer.pop(0)
            use_previous = False
        else:
            try:
                if not uncertain_pairs:
                    uncertain_pairs = deduper.uncertain_pairs()
                while True:
                    record_pair = uncertain_pairs.pop()
                    if (
                        len(
                            set([x["uri"] for x in record_pair]).intersection(
                                mapped_items
                            )
                        )
                        > 0
                    ):
                        examples_buffer.insert(0, (record_pair, "distinct"))
                        # TODO: do i need to process these?
                    else:
                        break
            except IndexError:
                break

        n_match = len(deduper.training_pairs["match"]) + sum(
            label == "match" for _, label in examples_buffer
        )
        n_distinct = len(deduper.training_pairs["distinct"]) + sum(
            label == "distinct" for _, label in examples_buffer
        )

        for pair in record_pair:
            for field in fields:
                line = "%s : %s" % (field, pair[field])
                print(line, file=sys.stderr)
            print(file=sys.stderr)

        print(
            "{0}/10 positive, {1}/10 negative".format(n_match, n_distinct),
            file=sys.stderr,
        )
        print(
            Fore.YELLOW + "Do these records refer to the same thing?" + Style.RESET_ALL,
            file=sys.stderr,
        )

        valid_response = False
        user_input = ""
        while not valid_response:
            if examples_buffer:
                prompt = "(y)es / (n)o / (u)nsure / (f)inished / (p)revious"
                valid_responses = {"y", "n", "u", "f", "p"}
            else:
                prompt = "(y)es / (n)o / (u)nsure / (f)inished"
                valid_responses = {"y", "n", "u", "f"}

            print(Fore.YELLOW + prompt + Style.RESET_ALL, file=sys.stderr)
            user_input = input()
            if user_input in valid_responses:
                valid_response = True

        if user_input == "y":
            examples_buffer.insert(0, (record_pair, "match"))
            mapped_items.add(record_pair[0]["uri"])
            mapped_items.add(record_pair[1]["uri"])
            confirmed_matches.append((record_pair[0]["uri"], record_pair[1]["uri"]))
            # deduper.mark_pairs({'match': record_pair})
        elif user_input == "n":
            examples_buffer.insert(0, (record_pair, "distinct"))
        elif user_input == "u":
            examples_buffer.insert(0, (record_pair, "uncertain"))
        elif user_input == "f":
            print(Fore.GREEN + "Finished labeling" + Style.RESET_ALL, file=sys.stderr)
            finished = True
        elif user_input == "p":
            use_previous = True
            uncertain_pairs.append(record_pair)

        if len(examples_buffer) > buffer_len:
            record_pair, label = examples_buffer.pop()
            if label in {"distinct", "match"}:

                examples: TrainingData
                examples = {"distinct": [], "match": []}
                examples[label].append(record_pair)  # type: ignore
                deduper.mark_pairs(examples)

    for record_pair, label in examples_buffer:
        if label in ["distinct", "match"]:

            examples: TrainingData
            examples = {"distinct": [], "match": []}
            examples[label].append(record_pair)  # type: ignore
            deduper.mark_pairs(examples)
    return confirmed_matches



from brickschema.graph import Graph
import rdflib
import pkgutil
import io


def test_simplify(owlrl_inference_backend):
    g = Graph(load_brick=True)
    data = pkgutil.get_data(__name__, "data/test.ttl").decode()
    g.load_file(source=io.StringIO(data))

    g.expand("owlrl", simplify=False, backend=owlrl_inference_backend)
    g.serialize("/tmp/test.ttl", format="ttl")

    q = "SELECT ?type WHERE { bldg:VAV2-4.ZN_T a ?type }"
    rows = list(g.query(q))
    bnodes = [r[0] for r in rows if isinstance(r[0], rdflib.BNode)]
    assert len(bnodes) >= 0

    g.simplify()

    rows = list(g.query(q))
    bnodes = [r[0] for r in rows if isinstance(r[0], rdflib.BNode)]
    assert len(bnodes) == 0



from .graph import Graph
from .namespaces import BRICK, A, REF, BACNET
from rdflib import Namespace, Literal
import BAC0
import logging
from typing import Optional

logging.basicConfig(level=logging.INFO)
logging.getLogger("BAC0_Root.BAC0").propagate = False
logging.getLogger("BAC0_Root.BAC0").setLevel(logging.WARNING)

NET = Namespace("urn:bacnet-scan/")


def clean_name(name):
    return name.replace(" ", "_").replace("-", "_").replace(".", "_").replace("/", "_")


# TODO: provide namespace for graph
def scan(ns: Namespace = NET, ip: Optional[str] = None) -> Graph:
    """
    Scan for BACnet devices on the provided network. If no network is provided,
    use the default scan logic which scans all available interfaces. Provide a
    network by indicating the IP address that BAC0 should bind to

    The Optional 'ns' parameter provides a namespace for the graph containing
    the scanned results.
    """
    graph = Graph()

    # ping=False avoids spamming messages on the network
    client = BAC0.connect(ip=ip, ping=False)
    client.discover()
    for dev in client.devices:
        print(dev)
        name, _, address, deviceid = dev
        name = clean_name(name)
        graph.add((ns[name], A, BACNET.BACnetDevice))
        graph.add((ns[name], BACNET["device-instance"], Literal(deviceid)))
        graph.add((ns[name], BACNET["hasAddress"], Literal(address)))

        logging.info(f"Scanning BACnet device {dev}")
        device = BAC0.device(
            dev[2], dev[3], client, history_size=0, segmentation_supported=False
        )
        for point in device.points:
            print(point)
            objectIdent = point.properties.address
            objectIRI = ns[name + "/" + str(objectIdent)]
            graph.add((objectIRI, A, BRICK.Point))
            props = [
                (A, REF.BACnetReference),
                (BACNET["object-identifier"], Literal(int(objectIdent))),
                (BACNET["objectOf"], ns[name]),
            ]
            props.append(
                (BACNET["object-name"], Literal(point.properties.name.strip()))
            )
            props.append(
                (
                    BACNET["object-description"],
                    Literal(point.properties.description.strip()),
                )
            )
            props.append((BACNET["object-type"], Literal(point.properties.type)))
            props.append((BACNET["units"], Literal(point.properties.units_state)))
            graph.add((objectIRI, REF.hasExternalReference, props))

    return graph



"""
The `main` module provides the CLI tool.
"""

from pathlib import Path
from warnings import warn

try:
    import typer

    from brickschema.brickify.src.handlers.Handler.Handler import Handler
    from brickschema.brickify.src.handlers.Handler.HaystackHandler.HaystackHandler import (
        HaystackHandler,
    )
    from brickschema.brickify.src.handlers.Handler.RACHandler.RACHandler import (
        RACHandler,
    )
    from brickschema.brickify.src.handlers.Handler.TableHandler import TableHandler
    from brickschema.brickify.util import minify_graph
except ImportError:
    warn(
        "brickschema needs to be installed with the 'brickify' option:\n\n\tpip install brickschema[brickify]"
    )
    import sys

    sys.exit(1)

app = typer.Typer()


@app.command(no_args_is_help=True)
def convert(
    source: str = typer.Argument(..., help="Path/URL to the source file"),
    input_type: str = typer.Option(
        help="Supported input types: rac, table, rdf, haystack-v4", default=None
    ),
    brick: Path = typer.Option(help="Brick.ttl", default=None),
    config: Path = typer.Option(help="Custom configuration file", default=None),
    output: Path = typer.Option(help="Path to the output file", default=None),
    serialization_format: str = typer.Option(
        help="Supported serialization formats: turtle, xml, n3, nt, pretty-xml, trix, trig and nquads",
        default="turtle",
    ),
    minify: bool = typer.Option(help="Remove inferable triples", default=False),
    input_format: str = typer.Option(
        help="Supported input formats: xls, csv, tsv, url, turtle, xml, n3, nt, pretty-xml, trix, trig and nquads",
        default="turtle",
    ),
    building_prefix: str = typer.Option(
        help="Prefix for the building namespace", default="bldg"
    ),
    building_namespace: str = typer.Option(
        help="The building namespace", default="https://example.com/bldg#"
    ),
    site_prefix: str = typer.Option(
        help="Prefix for the site namespace", default="site"
    ),
    site_namespace: str = typer.Option(
        help="The site namespace", default="https://example.com/site#"
    ),
):
    """
    The CLI tool uses this function to convert an input file to a Brick graph
    based on the file type and template.

    :param source: Path/URL to the source file
    :param input_type: Input file type. Supported input types: rac, table, rdf, haystack-v4
    :param brick: Path to Brick.ttl (used for minification)
    :param config: Path to custom configuration file
    :param output: Path to the output file
    :param serialization_format: Output graph format. Supported serialization formats: (turtle), xml, n3, nt, pretty-xml, trix, trig and nquads
    :param minify: Remove inferable triples?
    :param input_format: Supported input formats: xls, csv, tsv, url, turtle, xml, n3, nt, pretty-xml, trix, trig and nquads
    :param building_prefix: Prefix for the building namespace
    :param building_namespace: The building namespace
    :param site_prefix: Prefix for the site namespace
    :param site_namespace: The site namespace
    """
    if input_type in ["rac"]:
        handler = RACHandler
    elif input_type in ["haystack", "haystack-v4"]:
        handler = HaystackHandler
    elif input_type in ["table", "csv", "tsv"]:
        if input_type == "tsv":
            input_format = "tsv"
        handler = TableHandler
    elif input_type in [None, "rdf", "graph"]:
        handler = Handler
    else:
        message_start = typer.style("[Error] Input type: ", fg=typer.colors.RED)
        filename = typer.style(f"{input_type}", fg=typer.colors.RED, bold=True)
        message_end = typer.style(" not supported!", fg=typer.colors.RED)
        typer.echo(message_start + filename + message_end)
        raise typer.Exit(code=1)

    if handler:
        graph = handler(
            source=source,
            input_format=input_format,
            config_file=config,
        ).convert(building_prefix, building_namespace, site_prefix, site_namespace)

    minify_confirmed = None
    if minify is None:
        minify_confirmed = typer.confirm(
            "Do you want to remove inferable triples?", default=True
        )
    if minify or minify_confirmed:
        minify_graph(graph, brick)
    if not output:
        if source.startswith("http"):
            output = Path(source).name + ".brick.ttl"
        else:
            output = Path(source + ".brick.ttl")
    graph.serialize(destination=str(output), format=serialization_format)


if __name__ == "__main__":
    app()



from brickschema.inference import VBISTagInferenceSession
from brickschema.graph import Graph
from brickschema.namespaces import BRICK
from rdflib import Namespace
import pytest
import pkgutil
import io


def test_vbis_to_brick_inference():
    session = VBISTagInferenceSession()
    assert session is not None

    # input a fully-qualified VBIS tag, get Brick classes out
    test_cases = [
        ("ME-AHU-Su", BRICK.AHU),
        ("ME-AHU-Su-BU", BRICK.AHU),
        ("ME-ATU-VAV-SD", BRICK.VAV),
    ]
    for (vbistag, brickclass) in test_cases:
        predicted_classes = session.lookup_brick_class(vbistag)
        assert brickclass in predicted_classes


@pytest.mark.skip(
    reason="VBIS/Brick classification differences mean that this test does not pass currently"
)
def test_brick_to_vbis_inference_with_owlrl():
    ALIGN = Namespace("https://brickschema.org/schema/Brick/alignments/vbis#")

    # input brick model; instances should have appropriate VBIS tags
    g = Graph(load_brick=True)
    data = pkgutil.get_data(__name__, "data/vbis_inference_test.ttl").decode()
    g.load_file(source=io.StringIO(data))
    g.expand(profile="owlrl")
    g.expand(profile="vbis")

    test_cases = [
        ("http://bldg#f1", "ME-Fa"),
        ("http://bldg#rtu1", "ME-ACU"),
    ]
    for (entity, vbistag) in test_cases:
        query = f"SELECT ?tag WHERE {{ <{entity}> <{ALIGN.hasVBISTag}> ?tag }}"
        res = list(g.query(query))
        assert len(res) == 1
        assert str(res[0][0]) == vbistag

    conforms, _, results = g.validate(
        default_brick_shapes=True, min_iterations=1, max_iterations=1
    )
    assert conforms, results


# TODO: do without owlrl inference



import pytest
from brickschema.inference import (
    TagInferenceSession,
    HaystackInferenceSession,
    VBISTagInferenceSession,
)
from brickschema.namespaces import RDF, RDFS, BRICK, TAG, OWL
from brickschema.graph import Graph
from rdflib import Namespace, BNode
import io
import json
import pkgutil


def filter_bnodes(input_res):
    res = []
    for row in input_res:
        row = tuple(filter(lambda x: not isinstance(x, BNode), row))
        res.append(row)
    return list(filter(lambda x: len(x) > 0, res))


def test_tagset_inference(shacl_engine):

    g = Graph(load_brick=False)
    g.load_extension("shacl_tag_inference")
    data = pkgutil.get_data(__name__, "data/tags.ttl").decode()
    g.load_file(source=io.StringIO(data))
    g.compile(engine=shacl_engine)

    afs1 = g.query("SELECT ?x WHERE { ?x rdf:type brick:Air_Flow_Sensor }")
    assert len(afs1) == 1
    afsp1 = g.query("SELECT ?x WHERE { ?x rdf:type brick:Air_Flow_Setpoint }")
    assert len(afsp1) == 1
    mafs1 = g.query("SELECT ?x WHERE { ?x rdf:type brick:Max_Air_Flow_Setpoint_Limit }")
    assert len(mafs1) == 1


def test_lookup_tagset():
    session = TagInferenceSession(approximate=False)
    assert session is not None

    tagsets = session.lookup_tagset(["AHU", "Equipment"])
    assert tagsets[0][0] == set(["AHU"])

    tagsets = session.lookup_tagset(["Air", "Flow", "Sensor", "Point"])
    assert tagsets[0][0] == set(["Air_Flow_Sensor"])

    tagsets = session.lookup_tagset(["Air", "Flow", "Sensor", "Equipment"])
    assert len(tagsets) == 0

    tagsets = session.lookup_tagset(["Air", "Flow", "Setpoint", "Point"])
    assert tagsets[0][0] == set(["Air_Flow_Setpoint"])

    tagsets = session.lookup_tagset(
        ["Air", "Flow", "Setpoint", "Limit", "Parameter", "Point"]
    )
    assert tagsets[0][0] == set(["Air_Flow_Setpoint_Limit"])

    tagsets = session.lookup_tagset(
        ["Max", "Air", "Flow", "Setpoint", "Limit", "Parameter", "Point"]
    )
    assert tagsets[0][0] == set(["Max_Air_Flow_Setpoint_Limit"])


def test_most_likely_tagsets():
    session = TagInferenceSession(approximate=True)
    assert session is not None

    tagset1 = ["AHU", "Equipment"]
    inferred, leftover = session.most_likely_tagsets(tagset1)
    assert inferred == ["AHU"]
    assert len(leftover) == 0

    tagset2 = ["Air", "Flow", "Sensor"]
    inferred, leftover = session.most_likely_tagsets(tagset2)
    assert inferred == ["Air_Flow_Sensor"]
    assert len(leftover) == 0

    tagset3 = ["Air", "Flow", "Sensor", "Equipment"]
    inferred, leftover = session.most_likely_tagsets(tagset3)
    assert inferred == ["Air_Flow_Sensor"]
    assert len(leftover) == 1

    tagset4 = ["Air", "Flow", "Setpoint"]
    inferred, leftover = session.most_likely_tagsets(tagset4, num=1)
    assert inferred == ["Air_Flow_Setpoint"]
    assert len(leftover) == 0

    tagset5 = ["Air", "Flow", "Setpoint", "Limit"]
    inferred, leftover = session.most_likely_tagsets(tagset5, num=1)
    assert inferred == ["Air_Flow_Setpoint_Limit"]
    assert len(leftover) == 0

    tagset6 = ["Max", "Air", "Flow", "Setpoint", "Limit"]
    inferred, leftover = session.most_likely_tagsets(tagset6, num=1)
    assert inferred == ["Max_Air_Flow_Setpoint_Limit"]
    assert len(leftover) == 0


def test_brick_inference(shacl_engine):
    g = Graph(load_brick=True)
    g.load_extension("shacl_tag_inference")

    data = pkgutil.get_data(__name__, "data/brick_inference_test.ttl").decode()
    g.load_file(source=io.StringIO(data))

    g.compile(engine=shacl_engine)

    r = g.query("SELECT ?x WHERE { ?x rdf:type brick:Air_Temperature_Sensor }")
    urls = set([str(row[0]) for row in r])
    real_sensors = set(
        [
            "http://example.com/mybuilding#sensor1",
            "http://example.com/mybuilding#sensor2",
            "http://example.com/mybuilding#sensor3",
        ]
    )
    assert urls == real_sensors


@pytest.mark.skip(
    "Haystack inference is currently broken. See https://github.com/gtfierro/Brick-Haystack-harmonization"
)
def test_haystack_inference():
    data = pkgutil.get_data(__name__, "data/carytown.json").decode()
    raw_model = json.loads(data)
    brick_model = Graph(load_brick=True).from_haystack(
        "http://example.org/carytown", raw_model
    )
    points = brick_model.query(
        """SELECT ?p WHERE {
        ?p rdf:type/rdfs:subClassOf* brick:Point
    }"""
    )
    assert len(points) == 17

    equips = brick_model.query(
        """SELECT ?e WHERE {
        ?e rdf:type/rdfs:subClassOf* brick:Equipment
    }"""
    )
    assert len(equips) == 4


def test_rdfs_inference_subclass():
    EX = Namespace("http://example.com/building#")
    graph = Graph(load_brick=True).from_triples(
        [(EX["a"], RDF.type, BRICK.Temperature_Sensor)]
    )
    graph.expand(profile="rdfs")

    res1 = graph.query(
        f"""SELECT ?type WHERE {{
        <{EX["a"]}> rdf:type ?type
    }}"""
    )

    expected = [
        BRICK.Point,
        BRICK.Class,
        BRICK.Entity,
        BRICK.Sensor,
        RDFS.Resource,
        BRICK.Temperature_Sensor,
    ]

    # filter out BNodes
    res = []
    for row in res1:
        row = tuple(filter(lambda x: not isinstance(x, BNode), row))
        res.append(row)
    res = list(filter(lambda x: len(x) > 0, res))

    assert len(res) == len(expected), f"Results were {res}"
    for expected_class in expected:
        assert (expected_class,) in res, f"{expected_class} not found in {res}"


def test_inference_tags(shacl_engine):
    EX = Namespace("http://example.com/building#")
    graph = Graph(load_brick=True).from_triples(
        [(EX["a"], RDF.type, BRICK.Air_Flow_Setpoint)]
    )
    graph.compile(engine=shacl_engine)

    res1 = graph.query(
        f"""SELECT ?type WHERE {{
        <{EX["a"]}> rdf:type ?type
    }}"""
    )

    expected = [
        BRICK.Air_Flow_Setpoint,
    ]
    # filter out BNodes
    res1 = filter_bnodes(res1)

    assert set(res1) == set(map(lambda x: (x,), expected))

    res2 = graph.query(
        f"""SELECT ?tag WHERE {{
        <{EX["a"]}> brick:hasTag ?tag
    }}"""
    )

    expected = [
        TAG.Point,
        TAG.Air,
        TAG.Flow,
        TAG.Setpoint,
    ]
    res2 = filter_bnodes(res2)

    assert set(res2) == set(map(lambda x: (x,), expected))
